#ifndef TRACKREFINERPROCESSOR_H
#define TRACKREFINERPROCESSOR_H

#include "marlin/Processor.h"
#include "kaldet/GearTPCKalDetector.h"
#include "kaltest/TKalDetCradle.h"
#include "IMPL/TrackerHitImpl.h"

// include what you need from root
#include "TVector3.h"

// include what you need from std
#include <map>
#include <tuple>

namespace gear {
  class TPCModule;
}

namespace EVENT {
  class LCRunHeader;
  class LCEvent;
  class Track;
  class TrackerPulse;
}

namespace IMPL {
  class LCCollectionVec;
  class TrackImpl;
}

namespace marlintpc {

class RefinedTrack;

class TrackRefinerProcessor: public marlin::Processor {
	public:
		virtual Processor* newProcessor()
		{
			return new TrackRefinerProcessor;
		}

		TrackRefinerProcessor();

		virtual void init();

		virtual void processRunHeader(EVENT::LCRunHeader* run);

		virtual void processEvent(EVENT::LCEvent* evt);

		virtual void check(EVENT::LCEvent* evt);

		virtual void end();

		/// Utility functions
  		int getModuleID(const EVENT::TrackerHit* hitp);
  		int getRowID   (const EVENT::TrackerHit* hitp, const gear::TPCModule* tpcmodule);

	private:
		/// Copy TrackerHit.
		IMPL::TrackerHitImpl *copyTrackerHit(const IMPL::TrackerHitImpl *hitp);

		/// Copy Track.
		RefinedTrack         *copyTrack(const IMPL::TrackImpl                                             *trkp,
		                                const std::map<const EVENT::TrackerHit *, IMPL::TrackerHitImpl *> &theInputHitToOutputHitMap);

		/// Create KalHit from TrackerHit.
		kaldet::GearTPCHit   *createKalHit(const EVENT::TrackerHit *hitp);

		/// Create KalTrack from TrackerHitVec by Kalman Filtering it.
		TKalTrack *createKalTrack(const EVENT::TrackerHitVec                                   	  	&hitvec,
               		  		  const std::map<const EVENT::TrackerHit *, kaldet::GearTPCHit *>       &theInputHitToKalHitMap,
					  const bool                                                            dir);

		/// Calculate track's incident angle and expected hit position corresponding to the TrackerHit in question.
		std::tuple<double, TVector3>
			getIncidentPhiAndPosition(EVENT::Track                                                  *kaltrkp, 
			                    const EVENT::TrackerHit                                             *hitp, 
			                    const gear::TPCModule                                               *tpcmodule,
               		  		    const std::map<const EVENT::TrackerHit *, kaldet::GearTPCHit *>     &theInputHitToKalHitMap);


		/// Create a refined/split hit.
		IMPL::TrackerHitImpl *createRefinedHit(const TrackerHit                           *hitp,              // pointer to original hit
						             double                                r,                 // pad row radius
					                     double                                charge,            // charge
					                     double                                dcharge,           // charge error
					                     double                                rphi,              // r*phi
					                     double                                drphi,             // r*phi error
					                     std::map<int, EVENT::TrackerPulse *> &thePadToPulseMap); // pad ID to pulse map

		class MultiGaus {
			public:
				MultiGaus(int npar = 6) : _npar(npar) {};
				double operator()(double *x, double *par);

			private:
				int _npar;
		};

	protected:
		/* the place for protected and private member data and functions */
		std::string      _inputTrackerHitsColName;   ///< Name of the TPc pulse input collection
		std::string      _inputTrackColName;         ///< Name of the TPc pulse input collection
		std::string      _outputTrackerHitsColName;  ///< Name of the TPC hit output collection
		std::string      _outputTrackColName;        ///< Name of the track output collection
							     //
		double           _driftVelocity;             ///< drift velocity
		double           _maxDriftLength;            ///< maximum drift length
		double           _chargeConversionFactor;    ///< charge conversion factor
		std::vector<const gear::TPCModule *> _moduleVec;
		int                                  _nMod;

	private:
		TKalDetCradle         	      _kalDetCradle{};
		kaldet::GearTPCKalDetector   *_kalDetectorPtr{ nullptr };
};

} // namespace marlintpc
#endif // TRACKREFINERPROCESSOR_H
