#define KF_DEBUG
#include "TrackRefinerProcessor.h"
#include "RefinedTrack.h"

//std
#include <algorithm>
#include <iostream>
#include <string>
#include <cmath>
#include <map>
#include <tuple>
#include <list>

//MARLIN
#include "marlin/Global.h"

//GEAR
#include "gear/TPCParameters.h"
#include "gear/BField.h"
#include "gearimpl/Vector3D.h"
#include "gear/GEAR.h"
#include "gear/TPCModule.h"
#include "gear/TrackerPlanesParameters.h"
#include "gear/gearimpl/TPCModuleImpl.h"
#include "gearxml/GearXML.h"

// include what you need from lcio
#include "lcio.h"
#include "EVENT/LCRunHeader.h"
#include "EVENT/LCEvent.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackerHitImpl.h"
#include "IMPL/TrackerPulseImpl.h"
#include "IMPL/TrackImpl.h"
#include "IMPL/LCFlagImpl.h"
#include "Exceptions.h"

// KalTest
#include "kaltest/TKalTrackSite.h"
#include "kaltest/THelicalTrack.h"
#include "kaltest/TKalTrack.h"

// KalDetecotr, MeasLayer and KalHit
#include "kaldet/GearTPCKalDetector.h"
#include "kaldet/GearTPCCylinderMeasLayer.h"
#include "kaldet/GearTPCCylinderHit.h"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// ROOT
#include "TF1.h"
#include "TH1.h"
#include "TMath.h"
#include "TVector2.h"
#include "TVector3.h"
#include "TObjArray.h"
//#include "TFitResultPtr.h"
#include "TROOT.h"


using namespace lcio;
using namespace marlin;
using namespace std;
using namespace gear;
using namespace kaldet;

static const bool gkDir = kIterBackward ;

namespace marlintpc {

	TrackRefinerProcessor aTrackRefinerProcessor;

	/// Construct processor.
	TrackRefinerProcessor::TrackRefinerProcessor()
	      :	Processor("TrackRefinerProcessor")
	{
		// the processor description
		_description = "TrackRefinerProcessor is intended to refine TPC tracks";

		registerInputCollection(LCIO::TRACKERHIT, "InputTrackerHits", "Name of the input TrackerHits collection",
				_inputTrackerHitsColName, string("TPCHits"));

		registerInputCollection(LCIO::TRACK, "InputTracks", "Name of the input Track Collection", _inputTrackColName,
				string("TPCTracks"));

		registerOutputCollection(LCIO::TRACKERHIT, "OutputTrackerHits", "Name of the output TrackerHits collection",
				_outputTrackerHitsColName, string("RefinedTPCHits"));

		registerOutputCollection(LCIO::TRACK, "OutputTracks", "Name of the output Track Collection", _outputTrackColName,
				string("RefinedTPCTracks"));

		registerOptionalParameter("DriftVelocity", "Drift velocity", _driftVelocity, double(74.0));

		registerOptionalParameter("ChargeConversionFactor" ,
				"Optional: The charge conversion factor from ADC values to primary electrons",
				_chargeConversionFactor, double(1.)) ;
	}

	/// Initialize processor.
	void TrackRefinerProcessor::init() 
	{
		streamlog_out(DEBUG) << " init() called  " << endl;

		// Get gear information
		const gear::TPCParameters &theTPCParameters = marlin::Global::GEAR->getTPCParameters();
		_nMod = theTPCParameters.getNModules();
		for (int i_mod = 0; i_mod < _nMod ; i_mod++) {
			_moduleVec.push_back( &theTPCParameters.getModule( i_mod ) );
		}

		_maxDriftLength = theTPCParameters.getMaxDriftLength();  //20221014 change

    		// We have to register the kalDetector to the cradle.
    		// Afterwards we close the cradle (no more detectors can be added, we just have this one)
    		// and sort it so all the Kalman layers are in the right order according to their sorting 
    		// policy in the cradle.
    		// After this the order in the cradle and the kalDetector are not necessarily the same any more.
		_kalDetectorPtr = new GearTPCKalDetector(*Global::GEAR);
    		_kalDetCradle.Install(*_kalDetectorPtr);
    		_kalDetCradle.Close();
    		_kalDetCradle.Sort();
	}

	void TrackRefinerProcessor::processRunHeader(LCRunHeader*  /*run*/)
	{
		// do something for each run
	}

	/// Process event.
	void TrackRefinerProcessor::processEvent(LCEvent* evt) 
	{
		streamlog_out(DEBUG) << endl
				     << "===============================================================" << endl
				     << "Event Number: " << evt->getEventNumber() 			  << endl 
				     << "===============================================================" << endl
				     << flush;

		//--
		// Get input collections.
		//--
		// get input track collection 
		EVENT::LCCollection* colTracks = 0;
		try {
			colTracks = evt->getCollection( _inputTrackColName );
		} catch ( std::exception& ) {
			//std::cerr << "This event has no Track collection ! Skipped." << std::endl;
			return;
		}

		// get input hit collection 
		EVENT::LCCollection* colHits = 0;
		try {
			colHits = evt->getCollection( _inputTrackerHitsColName );
		} catch ( std::exception& ) {
			//std::cerr << "This event has no Track collection ! Skipped." << std::endl;
			return;
		}

		//--
		// Prepare output collections.
		//--
		// refined hit collection
		auto theOutputHitCol = new LCCollectionVec(LCIO::TRACKERHIT);
		// tpcHitCollection->setTransient(!_outputHitsPersistent);
		// set the flags that cellID1 should be stored
		LCFlagImpl hitFlag(0);
		hitFlag.setBit(LCIO::RTHBIT_ID1);
		theOutputHitCol->setFlag(hitFlag.getFlag());

		// refined track collection
		auto theOutputTrackCol = new LCCollectionVec(LCIO::TRACK);
		LCFlagImpl trkFlag(0);
		trkFlag.setBit(LCIO::TRBIT_HITS);
		theOutputTrackCol->setFlag(trkFlag.getFlag());

		//--
		// Prepare KalHits for KalTest, while copying input TrackerHit's to output TrackerHit's.
		//--
		map<const EVENT::TrackerHit *, GearTPCHit *>           theInputHitToKalHitMap;    // input TrackerHit to KalHit map
		map<const EVENT::TrackerHit *, IMPL::TrackerHitImpl *> theInputHitToOutputHitMap; // input TrackerHit to output TrackerHit map

		// Loop over input trackerHits to create KalHits, while copying them to theOutputHitCol.
		auto nHits = colHits->getNumberOfElements();

		streamlog_out(DEBUG) << "create KalHits: nHits = " << nHits << endl << flush;

		TObjArray kalhits;
		kalhits.SetOwner(); // this makes sure to clean up kalhits upon exiting this function

		for (int i_hit = 0; i_hit < nHits; i_hit++) {
			auto inhitp = static_cast<const IMPL::TrackerHitImpl *>(colHits->getElementAt(i_hit)) ;
			auto hitp   = copyTrackerHit(inhitp);
			theOutputHitCol->addElement(hitp);

			// update input TrackerHit to output TrackerHit map
			theInputHitToOutputHitMap.emplace(inhitp, hitp);

			// create KalHit
			auto kalhitp = createKalHit(hitp);

			// save it
			kalhits.Add(kalhitp);

			// and update TrackerHit to KalHit map
			theInputHitToKalHitMap.emplace(hitp, kalhitp);
		}

		//--
		// Copy input Track's to output Track's.
		//--
		auto nTrks = colTracks->getNumberOfElements();

		streamlog_out(DEBUG) << "Copy input tracks: nTrks = " << nTrks << endl << flush;

		for ( int i_trk = 0 ; i_trk < nTrks ; i_trk++ ) {
			auto  intrkp = static_cast<const IMPL::TrackImpl *>(colTracks->getElementAt(i_trk)) ;
			auto  trkp   = new RefinedTrack(intrkp, theInputHitToOutputHitMap);
			theOutputTrackCol->addElement(trkp);
		}

		//--
		// Loop over tracks and create TrackerHit to Track multi-map to detect overlapping hits.
		//--
		multimap<EVENT::TrackerHit *, RefinedTrack *> theHitToTrackMap;

		for (int i_trk = 0 ; i_trk < nTrks ; i_trk++) { // track loop
			auto trkp = static_cast<RefinedTrack *>(theOutputTrackCol->getElementAt(i_trk)) ;

			auto &hitvec = trkp->getTrackerHits();
                        for (auto hitp : hitvec) { // hit loop
				theHitToTrackMap.emplace(hitp, trkp);
			}
		}

		//--
		// Now look for hits shared by multiple tracks and split them into multiple hits by multi-gaussian fittting.
		//--
		//map<pair<EVENT::Track *, EVENT::TrackerHit *>, EVENT::TrackerHit *> theOldHitToRefinedHitMap;
		vector<RefinedTrack *> theModifiedTracks;

		EVENT::TrackerHit *lasthitp = nullptr;
		for (auto itr = theHitToTrackMap.begin(); itr != theHitToTrackMap.end(); ++itr) { // hit loop
			auto hitp  = itr->first; // hit address

			// skip repeated hits
			if (hitp == lasthitp) continue;
			else                  lasthitp = hitp;

			auto modID = getModuleID(hitp) ; 
			auto rowID = getRowID(hitp, _moduleVec[modID]);
			auto count = theHitToTrackMap.count(hitp);      // # overlapping tracks	

			//--
			// Find overlapping hits.
			//--
			if (count > 1) { // found an overlapping hit 

				streamlog_out(DEBUG) << endl
						     << "-------------------------------------------------------" << endl
						     << "found an overlapping hit: " << static_cast<void *>(hitp) 
						     << "  count = " << count 		  			  << endl
						     << "-------------------------------------------------------" << endl
						     << flush;
				//--
				// Preparation for pad response function width calculation using 
				//      sigma_PR^2 = sigma_PRF^2 + Cd^2 * dl + (l^2/12) * tan^2(phi)
				// where
				// 	sigma_PRF: intrinsic pad response function width
				// 	Cd	 : transverse diffusion constant
				// 	dl	 : drift length
				// 	l	 : pad length
				// 	phi	 : incident phi angle
				//--
				// FIXME /////////////////////////////////////////////////////
				//     the following should be moved to more appropriate place
				static const double kSqrt2Pi    = sqrt(2.*M_PI);
				static const double kSigma_PRF  = 0.4828;  // [mm]
				static const double kSigma2_PRF = kSigma_PRF * kSigma_PRF;
				static const double kCd         = 0.09074; // [mm/sqrt(cm)] 
				static const double kCd2        = kCd * kCd;
				//////////////////////////////////////////////////////////////

				auto PadHeight = _moduleVec[modID]->getRowHeight(rowID);
				auto kL        = PadHeight;     // [mm]
				auto kL2by12   = kL * kL / 12.;

				auto pos       = hitp->getPosition(); // in 3d global frame
				auto side      = pos[2] > 0 ? +1 : -1;
				auto dl        = std::max(_maxDriftLength - side * pos[2], 0.) * 0.1;     // drift length[cm]

				//--
				// loop over tracks to decide sigma_PR and expected rphi coordinate
				// and store results in theTrackToSigmaPRAndRphiMap.
				//--
				auto range = theHitToTrackMap.equal_range(hitp);
				map<EVENT::Track *, tuple<double, double, double>> theTrackToSigmaPRAndRphiMap; // (sigma, r, rphi)_exp
				map<int, RefinedTrack *>                           theTrackNoToTrackMap;        // (track No., Track*)
				int ntrk = 0;
				for (auto it = range.first; it != range.second; ++it) { // track loop
					auto trkp = it->second;
					auto [phiinc, xvexpect] = getIncidentPhiAndPosition(trkp, 
											    hitp, 
											    _moduleVec[modID],
										 	    theInputHitToKalHitMap);
#ifdef KF_DEBUG
					streamlog_out(DEBUG) << endl 
							     << "calculate incident phi angle for track No:" << ntrk << endl
							     << "    phiinc: " << phiinc  			     << endl
							     << flush;
#endif
					// calculate PRF width from sigma_PR^2 = sigma_PRF^2 + Cd^2 * dl + (l^2/12) * tan^2(phi)
					
					auto sigma_PR  =  sqrt(kSigma2_PRF + kCd2 * dl + kL2by12 * pow(tan(phiinc), 2));

					auto rphivexploc = _moduleVec[modID]->globalToLocal(xvexpect[0], xvexpect[1]); // global in Cartesian and local in polar
					auto rowRadius   = rphivexploc[0];
					auto phi         = rphivexploc[1];
					auto rphiexp     = rowRadius * phi;
					theTrackToSigmaPRAndRphiMap[trkp] = {sigma_PR, rowRadius, rphiexp};

					theTrackNoToTrackMap[ntrk] = trkp;
					ntrk++;
				}

				//--
				// loop over pulses and store (rphi, charge) for each pad in rhpi_vs_charge
				// as well as to prepare padID to TrackerPulse map.
				//--
				auto pulsesInHit = hitp->getRawHits();                // vector containing TrackerPulses*
				auto maxCharge   = -9999.;                            // peak pulse charge
                                auto padWidth    = 0.;                                // pad width
				vector<tuple<double, double, double>> rphi_vs_charge; // arrays of (rhpi, charge) per pad
				map<int, TrackerPulse *>      thePadToPulseMap;       // pad ID to pulse map
				for (auto &lcop : pulsesInHit) { // loop over pulses stored as LCObjects
					auto pulsep         = static_cast<TrackerPulse *>(lcop);
					auto padID          = pulsep->getCellID0();
					auto padCenter      = _moduleVec[modID]->getPadCenter(padID);
					auto localPadCenter = _moduleVec[modID]->globalToLocal(padCenter[0], padCenter[1]); // local in polar

					auto rowRadius      = localPadCenter[0];
					auto padPhi         = localPadCenter[1];
					auto rphi           = rowRadius * padPhi;

					auto charge         = pulsep->getCharge();
					auto dcharge        = sqrt(double((pulsep->getCovMatrix())[0]));
					if (charge > maxCharge) {
						maxCharge = charge;
						padWidth  = rowRadius * _moduleVec[modID]->getPadWidth(pulsep->getCellID0()); // gePadWidth in radian
					}
                                        rphi_vs_charge.push_back(tuple<double, double, double>{rphi, charge, dcharge});
					thePadToPulseMap.emplace(padID, pulsep);
				}

				//--
				// skip this overlapping hit if the number of pads is not enough.
				//--
				auto npads = rphi_vs_charge.size();
				if (npads < 2*count) {
					continue;
				}

				//--
				// histogram charge distribution of this overlapping hit for multi-gaussian fit.
				//--
				auto [rphiFirst, chargeFirst, dchargeFirst] = rphi_vs_charge[0];
				auto [rphiLast , chargeLast , dchargeLast ] = rphi_vs_charge[npads-1];
#if 0
				auto rphimin = rphiFirst - 1.5*padWidth;
				auto rphimax = rphiLast  + 1.5*padWidth;
				TH1D hist("cdist", "", rphi_vs_charge.size() + 2, rphimin, rphimax);
#else
				auto rphimin = rphiFirst - 0.5*padWidth;
				auto rphimax = rphiLast  + 0.5*padWidth;
				TH1D hist("cdist", "", rphi_vs_charge.size(), rphimin, rphimax);
#endif
#ifdef KF_DEBUG
				streamlog_out(DEBUG) << endl 
						     << "now split the overlapping hit : npads = " << npads << endl
						     << "    row ID   = "           << rowID
						     << "    padWidth = "           << padWidth 
						     << "    (rphimin,rphimax) = (" << rphimin << ", " << rphimax << ")" << endl
						     << "    --------------------------------------------"        	 << endl
						     << "    charge dist: ";
#endif
				for (auto [rphi, charge, dcharge] : rphi_vs_charge) {
					auto ibin = hist.FindBin(rphi);
					hist.SetBinContent(ibin, charge);
					hist.SetBinError  (ibin, dcharge);
#ifdef KF_DEBUG
					streamlog_out(DEBUG) << " " << charge;
#endif
				}
#ifdef KF_DEBUG
				streamlog_out(DEBUG) << endl
						     << "    --------------------------------------------" << endl << flush;
#endif

				//--
				// fit this set of TrackerPulses* with a multi-gaussian.
				//--
				auto npar = 3*count; // (peak height, mean, sigma) x # overlapping tracks
				MultiGaus multig(npar);

				TF1 fun("multigaus", multig, rphimin, rphimax, npar);

				// set initial parameters for multi-gaussian fit
				for (auto itrk = 0; itrk < int(count); itrk++) {
					auto trkp             = theTrackNoToTrackMap[itrk];
					auto [sigma, r, mean] = theTrackToSigmaPRAndRphiMap[trkp];
#ifdef KF_DEBUG
					streamlog_out(DEBUG) << "    set initial fit parameters: itrk = " << itrk 	     << endl
							     << "        sigma: " << sigma << " mean: " << mean << " r: " << r << endl
							     << flush;
#endif
					fun.SetParameter(3*itrk    , maxCharge/count); 
					fun.SetParameter(3*itrk + 1, mean);
					fun.FixParameter(3*itrk + 2, sigma);
					fun.SetParLimits(3*itrk + 1, rphimin, rphimax);
				}

				// now multi-gaussian fit the charge distribution to split this hit 
#ifdef KF_DEBUG
				hist.Fit(&fun, "I", "goff", rphimin, rphimax);
#else
				hist.Fit(&fun, "IQ", "goff", rphimin, rphimax);
#endif

				//--
				// loop over opverlapping tracks and create a new TrackerHit corresponding to each mean value of 
				// the resultant gaussians.
				//--
				for (auto itrk = 0; itrk < int(count); itrk++) { // loop over overlapping tracks contributing to this hit
					auto trkp  = theTrackNoToTrackMap[itrk];
					auto [sigma, r, rphiexp] = theTrackToSigmaPRAndRphiMap[trkp];

					// get multi-gaussian fit results
					auto charge  = fun.GetParameter(3*itrk) * kSqrt2Pi;
					auto dcharge = fun.GetParError (3*itrk) * kSqrt2Pi;
					auto rphi    = fun.GetParameter(3*itrk + 1);
					auto drphi   = fun.GetParError (3*itrk + 1);

					streamlog_out(DEBUG) << "-------------------------------------------" << endl
							     << "multi-gaussian fit results: itrk = " << itrk		<< endl
							     << "    charge: " << charge << " dcharge: " << dcharge 	<< endl
							     << "    rphi  : " << rphi   << " drphi  : " << drphi	<< endl
							     << flush;

					auto newHitp = createRefinedHit(hitp, r, charge, dcharge, rphi, drphi, thePadToPulseMap);

					// Note: 
		                        // 	Standard definitions:
                                        // 	int const CONTAINS_DEAD_CHANNEL_BIT     = 1;
                                        // 	int const CONTAINS_NOISY_CHANNEL_BIT    = 1 << 1;
                                        // 	int const HIT_AT_MODULE_BORDER_BIT      = 1 << 2;
                                        // 	int const CONTAINS_OVERFLOW_PULSE_BIT   = 1 << 3;
                                        // 	int const CONTAINS_SPLIT_PULSE_BIT      = 1 << 4;
                                        // 	int const CONTAINS_ANOMALOUS_PULSE_BIT  = 1 << 5;
                                        // 	int const HIT_NEXT_TO_DEAD_CHANNEL_BIT  = 1 << 6;
                                        // 	int const MULTIPLE_HIT_CANDIDATE_BIT    = 1 << 7;
                                        // 	int const REMOVE_FROM_TRACK_FINDING_BIT = 1 << 8;
                                        // 	int const CONTAINS_PLATEAU_CUTOFF_PULSE_BIT  = 1 << 9;
		                        //
		                        // 	User defined:
		                        // 	int const SPLIT_HIT_BIT                 = 1 << 20; // bit 20
					// 
					// FIXME: //////////////////////////////////////////////////////////
					//    to be moved to more appropriate place
					int const SPLIT_HIT_BIT_INDEX = 20; // bit 20
					////////////////////////////////////////////////////////////////////
					LCFlagImpl aflag(newHitp->getQuality());
					aflag.setBit(SPLIT_HIT_BIT_INDEX);
					auto qualityFlag = aflag.getFlag();

		                        newHitp->setQuality(qualityFlag);

					// save the refined hit in output collection
					theOutputHitCol->addElement(newHitp);

					// fill the map from old hit to new hit
					//theOldHitToRefinedHitMap.emplace(pair(trkp, hitp), newHitp);

					// replace the old hit with this new hit in the track
					// Note:
					// 	typedef std::vector<TrackerHit *> TrackerHitVec ;
					//
					auto &hitvec = trkp->trackerHits();
					auto iter = find(hitvec.begin(), hitvec.end(), hitp);
					*iter = newHitp;

					// Save this track in the modified track list.
					theModifiedTracks.push_back(trkp); 
				}
				//--
				// Mark the original hit as unusable.
				//--
				// Note:
		                // Standard definitions:
                                // 	int const CONTAINS_DEAD_CHANNEL_BIT     = 1;
                                // 	int const CONTAINS_NOISY_CHANNEL_BIT    = 1 << 1;
                                // 	int const HIT_AT_MODULE_BORDER_BIT      = 1 << 2;
                                // 	int const CONTAINS_OVERFLOW_PULSE_BIT   = 1 << 3;
                                // 	int const CONTAINS_SPLIT_PULSE_BIT      = 1 << 4;
                                // 	int const CONTAINS_ANOMALOUS_PULSE_BIT  = 1 << 5;
                                // 	int const HIT_NEXT_TO_DEAD_CHANNEL_BIT  = 1 << 6;
                                // 	int const MULTIPLE_HIT_CANDIDATE_BIT    = 1 << 7;
                                // 	int const REMOVE_FROM_TRACK_FINDING_BIT = 1 << 8;
                                // 	int const CONTAINS_PLATEAU_CUTOFF_PULSE_BIT  = 1 << 9;
				// 
				auto qualityFlag = hitp->getQuality();
				hitflag::setMultipleHitCandidate(qualityFlag);
				hitflag::setRemoveFromTrackFinding(qualityFlag);
		                static_cast<IMPL::TrackerHitImpl *>(hitp)->setQuality(qualityFlag);

			} // end of an overlapping hit
		} // end of hit loop

		//--
		// Modify output tracks by Kalman-refiltering refined hits.
		//--
		TObjArray kaltracks;
		kaltracks.SetOwner(); // this makes sure to clean up kaltracks upon exiting this function

		// remove duplication in theModifiedTracks
		sort(theModifiedTracks.begin(), theModifiedTracks.end());
		theModifiedTracks.erase(unique(theModifiedTracks.begin(), theModifiedTracks.end()), theModifiedTracks.end());
#ifdef KF_DEBUG
		// FIXME: //////////////////////////////////////////////////////////
		//    to be moved to more appropriate place
		int const SPLIT_HIT_BIT_INDEX = 20; // bit 20
		////////////////////////////////////////////////////////////////////
		auto trkNo = 0;
#endif
		for (auto trkp : theModifiedTracks) {
			auto hitvec = trkp->trackerHits();
#ifdef KF_DEBUG
			streamlog_out(DEBUG) << "--------------------------------" 		     << endl
					     << " trkNo " << trkNo << ": #hits = " << hitvec.size()  << endl
					     << flush;
#endif
			for (auto hitp : hitvec) {
#ifdef KF_DEBUG
				auto qualityFlag = hitp->getQuality();
				LCFlagImpl aflag(qualityFlag);
				auto ismultihit  = hitflag::isMultipleHitCandidate(qualityFlag);
				auto issplithit  = aflag.bitSet(SPLIT_HIT_BIT_INDEX);
				auto modID = getModuleID(hitp);
				auto rowID = getRowID(hitp, _moduleVec[modID]);
				streamlog_out(DEBUG) << "     Module: " << modID << " Row: " <<  rowID 
						     << " hitp: " << static_cast<const void *>(hitp)
						     << " multihit? " << (ismultihit ? "yes" : "no") 
						     << " splithit? " << (issplithit ? "yes" : "no") << endl
						     << flush;
#endif
				auto kalhitp = createKalHit(hitp);
				kalhits.Add(kalhitp);
				theInputHitToKalHitMap[hitp] = kalhitp;
			}
			auto kaltrkp = createKalTrack(hitvec, theInputHitToKalHitMap);

			if (kaltrkp) {
				kaltracks.Add(kaltrkp);

				trkp->modify(kaltrkp);
			} else {
				streamlog_out(DEBUG) << " Too many discarded hits, track not modified" << endl
						     << flush;
			}
#ifdef KF_DEBUG
			trkNo++;
#endif
		}

		streamlog_out(DEBUG) << endl 
				     << "-------------------------------------------" << endl
				     << "Kalman refiltering finished " 		      << endl 
				     << "-------------------------------------------" << endl
				     << flush;

		//--
		// Store refined tracks and hits in event buffer.
		//--
		evt->addCollection(theOutputHitCol  , _outputTrackerHitsColName);
		evt->addCollection(theOutputTrackCol, _outputTrackColName);
	}


	void TrackRefinerProcessor::check([[maybe_unused]] EVENT::LCEvent* evt)
	{
	}

	void TrackRefinerProcessor::end()
	{
	}


	/// Utility member functions
	int TrackRefinerProcessor::getModuleID( const TrackerHit* hitp ) 
	{

		TrackerPulse* plsp = static_cast< TrackerPulse* >( hitp->getRawHits()[0] ) ;

		int mod = plsp->getCellID1() ;

		return mod;

	}

	int TrackRefinerProcessor::getRowID( const TrackerHit* hitp , const gear::TPCModule* tpcmodule )
	{
		TrackerPulse* plsp = static_cast< TrackerPulse* >( hitp->getRawHits()[0] ) ;

		int padID = plsp->getCellID0() ;

		int rowID = tpcmodule->getRowNumber( padID ) ;

		return rowID ;
	}

	/// Copy TrackerHit.
	TrackerHitImpl *TrackRefinerProcessor::copyTrackerHit(const TrackerHitImpl *hitp)
	{
		auto newHitp = new TrackerHitImpl();

	        newHitp->setType(hitp->getType());

		newHitp->setCellID0(hitp->getCellID0());
		newHitp->setCellID1(hitp->getCellID1());

		newHitp->setPosition (hitp->getPosition());
		newHitp->setCovMatrix(hitp->getCovMatrix());

		newHitp->setEDep     (hitp->getEDep());
	        newHitp->setEDepError(hitp->getEDepError());

	        newHitp->setTime(hitp->getTime());

		auto pulsesInHit = hitp->getRawHits(); // vector containing TrackerPulses*
		for (auto &lcop : pulsesInHit) { // loop over pulses stored as LCObjects
			auto pulp = static_cast<TrackerPulse *>(lcop);
			newHitp->rawHits().push_back(pulp);
		}

		newHitp->setQuality(hitp->getQuality());

		return newHitp;
	}
	
	
	/// Create KalHit from TrackerHit for KalTest.
	GearTPCHit *TrackRefinerProcessor::createKalHit(const EVENT::TrackerHit *hitp)
	{
		auto modID = getModuleID(hitp) ; 
		auto rowID = getRowID(hitp, _moduleVec[ modID ]);
		auto hpos  = hitp->getPosition();       // global frame
		TVector3 xv(hpos[0], hpos[1], hpos[2]); // 3d hit position vector

		auto   measlayerPtr = _kalDetectorPtr->GetMeasLayer(modID, rowID);
		double meas [2];
		double dmeas[2];
		auto mv   = measlayerPtr->XvToMv(xv); // meas. vector (r*phi, z)
		auto side = hpos[2] > 0 ? +1 : -1;
		auto d    = _maxDriftLength - side * hpos[2]; // drift length
		meas [0]  = mv( 0 , 0 ) ;
		meas [1]  = mv( 1 , 0 );
		dmeas[0]  = measlayerPtr->GetSigmaX(d) ;
		dmeas[1]  = measlayerPtr->GetSigmaZ(d) ;

		auto qualityFlag = hitp->getQuality();
		LCFlagImpl aflag(qualityFlag);

		// FIXME: //////////////////////////////////////////////////////////
		//    to be moved to more appropriate place
		int const SPLIT_HIT_BIT_INDEX = 20; // bit 20
		////////////////////////////////////////////////////////////////////
		if (hitflag::isMultipleHitCandidate(qualityFlag)) { // FIXME: very temporary
			if (!aflag.bitSet(SPLIT_HIT_BIT_INDEX)) dmeas[0] *= 1.e1;
			else				  	dmeas[0] *= 2.;
		}

		auto b       = marlin::Global::GEAR->getBField().at(Vector3D(hpos[0], hpos[1], hpos[2]))[2];
		auto kalhitp = new GearTPCCylinderHit(*measlayerPtr, meas, dmeas, hitp, b, _driftVelocity);

		return kalhitp;
	}

	/// Create KalTrack from TrackerHitVec by Kalman Filtering it.
	TKalTrack *TrackRefinerProcessor::createKalTrack(const EVENT::TrackerHitVec                                  &hitvec,
               		  			 	 const map<const EVENT::TrackerHit *, GearTPCHit *>          &theInputHitToKalHitMap)
	{		
		// Note:
		//        typedef std::vector<TrackerHit *> TrackerHitVec;
		//--
		// Choose 3 points to decide starting helix.
		//--
		int i1 , i2 , i3 ; // (i1,i2,i3) = (1st,mid,last) hit to filter
		if (gkDir == kIterBackward) {
			i3 = 1 ; 
			i1 = hitvec.size() -2 ;
			i2 = i1 /2;
		} else {
			i1 = 1 ;
			i3 = hitvec.size() - 2 ; 
			i2 = i3 / 2 ;
		}

		//--
		// Create dummy site.
		//--
		auto  ht1Ptr = theInputHitToKalHitMap.at(hitvec[i1]);
		auto  htdPtr = new GearTPCCylinderHit(*static_cast<GearTPCCylinderHit *>(ht1Ptr)); // dummy hit
		auto &hitd   = *htdPtr;
		      hitd(0,1) = 1.e6; // huge error on r*phi
		      hitd(1,1) = 1.e6; // huge error on z
		auto &sited = *new TKalTrackSite( hitd );
		sited.SetHitOwner();
		sited.SetOwner();

		// decide starting helix using 3 points from kalhits
		auto &h1 = *static_cast<TVTrackHit *>(theInputHitToKalHitMap.at(hitvec[i1]));
		auto &h2 = *static_cast<TVTrackHit *>(theInputHitToKalHitMap.at(hitvec[i2]));
		auto &h3 = *static_cast<TVTrackHit *>(theInputHitToKalHitMap.at(hitvec[i3]));
		auto  x1 = h1.GetMeasLayer().HitToXv(h1);
		auto  x2 = h2.GetMeasLayer().HitToXv(h2);
		auto  x3 = h3.GetMeasLayer().HitToXv(h3);
		THelicalTrack helstart(x1, x2, x3, h1.GetBfield(), gkDir);

		// set dummy state vector
		static TKalMatrix svd(kSdim, 1);
		svd( 0 , 0 ) = 0.;
		svd( 1 , 0 ) = helstart.GetPhi0();
		svd( 2 , 0 ) = helstart.GetKappa();
		svd( 3 , 0 ) = 0.;
		svd( 4 , 0 ) = helstart.GetTanLambda();
		if ( kSdim == 6 ) svd(5 , 0) = 0.; // t0
		static TKalMatrix C(kSdim, kSdim);
		for (int i = 0; i < kSdim; i++) C(i,i) = 1.e4 ; // dummy error matrix
		sited.Add(new TKalTrackState(svd, C, sited, TVKalSite::kPredicted));
		sited.Add(new TKalTrackState(svd, C, sited, TVKalSite::kFiltered));

		// create KalTrack
		auto  kaltrkp  = new TKalTrack; // a track is a kal system
		auto &kaltrack = *kaltrkp;
		kaltrack.SetOwner();    // track owns sites
		kaltrack.Add( &sited ); // add dummy site

     		// Now start Kalman filter 
		for (auto hitp : hitvec) { // loop over trackerHit
			auto kalhitp = theInputHitToKalHitMap.at(hitp);    // corresponding kalhit
			auto &site   = *new TKalTrackSite(*kalhitp);

			if (!kaltrack.AddAndFilter(site)) {
				streamlog_out(DEBUG2) << "Kalman Filter: site discarded!" << std::endl;
				delete &site;
			}
		} // filterling ends here

		//--
		// Make sure that # discarded sites is small enough.
		//--
		auto frac = 1. - (kaltrack.GetEntries() - 1.)/hitvec.size();

		if (frac > 0.) {
			streamlog_out(DEBUG) << "createKalTrack: " << static_cast<void *>(kaltrkp) << " fraction discarded = " << frac << endl << flush;
		}

		// FIXME: //////////////////
		auto kMaxBadFracCut = 0.2;
		////////////////////////////
		if (frac > kMaxBadFracCut) {
			delete kaltrkp;
			kaltrkp = nullptr;
		}

		//--
		// That's it!
		//--
		return kaltrkp;
	}

	/// Calculate track's incident angle and expected hit position corresponding to the TrackerHit in question.
	tuple<double, TVector3> TrackRefinerProcessor::getIncidentPhiAndPosition(EVENT::Track                                          *trkp, 
			                                                   const EVENT::TrackerHit                                     *hitp, 
			                                                   const gear::TPCModule                                       *tpcmodule,
               		  			 	 		   const map<const EVENT::TrackerHit *, GearTPCHit *>          &theInputHitToKalHitMap)
	{
		// local utility lambda function
		auto toBaseRange = [](double phi)
		{
			while (phi <= -M_PI) {
				phi += 2. * M_PI ;
			}
			while (phi >   M_PI) {
				phi -= 2. * M_PI ;
			}
			return phi ;
		};

		// calculate radial vector(rv) at the hit point 
		auto pos   = hitp->getPosition();                               // in global frame (Cartesian)
		auto rphiv = tpcmodule->globalToLocal(pos[0], pos[1]); // in module-local frame (polar)
		auto r     = rphiv[0];
		auto phi   = rphiv[1];
		TVector2 rv2(r*cos(phi), r*sin(phi));

		// FIXME: this should be somewhere more appropriate ////////////////////////////
   		static const Double_t kLightVelocity = 2.99792458e8; //! light velocity [m/sec]
   		static const Double_t kGiga          = 1.0e9;        //! Giga = 10^{9}
		////////////////////////////////////////////////////////////////////////////////

		// then calculate tangential vector(tv) at the hit point 
		auto &kalhit = *theInputHitToKalHitMap.at(hitp);
		auto bz      =  kalhit.GetBfield();
    		auto phi0    =  toBaseRange(trkp->getPhi() - M_PI / 2.);
		auto ptor   =  kGiga/kLightVelocity*1000./bz; // units: mm, sec, Tesla
		auto kappa   =  ptor * trkp->getOmega();
    		auto dr      = -trkp->getD0();
    		auto dz      =  trkp->getZ0();
    		auto tanl    =  trkp->getTanLambda();
		auto refpos  =  trkp->getReferencePoint();
		auto x0      =  double(refpos[0]);
		auto y0      =  double(refpos[1]);
		auto z0      =  double(refpos[2]);

		THelicalTrack heltrack(dr, phi0, kappa, dz, tanl, x0, y0, z0, bz);

		TVector3 xvexpect(pos[0], pos[1], pos[2]);
		auto &measlyr  = kalhit.GetMeasLayer(); 				    // TVMeasLayer
		auto &surface  = dynamic_cast<const TCylinder &>(measlyr);		    // TCylinder
		auto  dphi     = 0.;							    // deflection angle from pivot
		auto  mode     = 0;
		heltrack.MoveTo(xvexpect, dphi);					    // move pivot to the hit point
		      dphi     = 0.;
		auto  nxp      = surface.CalcXingPointWith(heltrack, xvexpect, dphi, mode); // calculate expected hit point
                if (nxp != 0) {
			auto  globaltv = heltrack.CalcDxDphi(dphi);                                 // tangential vector in global frame (Cartesian)
			auto  rphitv   = tpcmodule->globalToLocal(globaltv(0,0), globaltv(1,0));    // tangential vector in module-local frame (polar)
                	auto  rtv      = rphitv[0];
                	auto  phitv    = rphitv[1];
			TVector2 tv2(rtv*cos(phitv), rtv*sin(phitv));                               // tangential vector in module-local frame (Cartesian)
			tv2 = tv2.Unit();

			if (kappa > 0) tv2 *= -1. ;

			// calculate incident angle(phi) as angle between tv2 and rv2
			dphi = rv2.DeltaPhi(tv2);
			dphi = toBaseRange(dphi);
		} else { // could not find any crossing point. 
			dphi = 0.;
			xvexpect.SetXYZ(pos[0], pos[1], pos[2]); //FIXME: do something more sensible here.

			streamlog_out(ERROR) << "found no crossing point" << endl;
		}

		return {dphi, xvexpect}; // (phi_inc, xv_expected)
	}


	/// Multi-gaussian fitting function
        double TrackRefinerProcessor::MultiGaus::operator()(double *xp, double *par)
	{
		auto sum = 0.;
		for (int itrk = 0; itrk < _npar/3; itrk++) {
			auto peak  = par[3*itrk  ];
			auto mean  = par[3*itrk+1];
			auto sigma = par[3*itrk+2];
			auto x = xp[0];
			sum += peak*exp(-0.5*pow((x-mean)/sigma, 2));
		}

		return sum;
	}

	/// Create a refined/split hit
	TrackerHitImpl *TrackRefinerProcessor::createRefinedHit(const TrackerHit         *hitp,              // pointer to original hit
								double                    r,                 // pad row radius
					                        double                    charge,            // charge
					                        double                    dcharge,           // charge error
					                        double                    rphi,              // r*phi
					                        double                    drphi,             // r*phi error
					                        map<int, TrackerPulse *> &thePadToPulseMap)  // pad ID to pulse map
	{
		auto   modID         = getModuleID(hitp) ; 
		auto   phi           = rphi/r;
		auto   xv2           = _moduleVec[modID]->localToGlobal(r, phi); // local in polar and global in Cartesian
		auto   pos           = hitp->getPosition(); // in 3d global frame
		double newPosition[] = {xv2[0], xv2[1], pos[2]}; // new hit position in global frame


		// find the nearest pad
		const gear::TPCParameters &theTPCParameters = marlin::Global::GEAR->getTPCParameters();
		auto  globalPadIndx = theTPCParameters.getNearestPad(newPosition[0], newPosition[1], newPosition[2]);
		auto  padID  = globalPadIndx.getPadIndex();
	        auto  pulsep = thePadToPulseMap[padID];;  // nearest pulse
#ifdef KF_DEBUG
		auto oldplsp 	    = static_cast<TrackerPulse *>(hitp->getRawHits()[0]);
		auto oldPadID       = oldplsp->getCellID0();
		auto oldRowID       = getRowID(hitp, _moduleVec[modID]);
		auto rowID          = _moduleVec[modID]->getRowNumber(padID);

		streamlog_out(DEBUG) << endl
				     << " createRefinedHit: local  (x, y) = (" << r*cos(phi) << ", " << r*sin(phi) << ")" << endl
		                     << "                 : global (x, y) = (" << xv2[0] << ", " << xv2[1] << ")"         << endl
		                     << "             old : global (x, y) = (" << pos[0] << ", " << pos[1] << ")"         << endl
				     << "                 : radius = "  << r << " phi = " << phi                           << endl
				     << "                 : modId  = "  << modID
				     << "       (old,new) : rowId  = (" << oldRowID << ", " << rowID << ")"               << endl
				     << "       (old,new) : padId  = (" << oldPadID << ", " << padID << ")"               << endl
				     << "                 : pulsep = "  << static_cast<void *>(pulsep)                    << endl
				     << flush;
#endif
		auto z  = pos[2];
		auto dz = sqrt(hitp->getCovMatrix()[5]);
		if (pulsep) {
			// z from pulse on the center pad
			auto  dlen  = _driftVelocity * pulsep->getTime()*1.e-3; // time in [ns]
			auto  ddlen = pow(_driftVelocity, 2) * (pulsep->getCovMatrix()).at(2)*1.e-6; // [mm^2]
			auto  side  = pos[2] > 0 ? +1 : -1;
	        	      z     = _maxDriftLength - side * dlen;
			      dz    = ddlen;
		}
#ifdef KF_DEBUG
		streamlog_out(DEBUG) << "                 : global z = " << z << " +/- " << dz << endl << flush;
#endif
		newPosition[2] = z;

		auto newHitp = new TrackerHitImpl();

		// set cell IDs
		newHitp->setCellID0(modID);
		newHitp->setCellID1(padID);

		//--
		// covMatrix[] = { <dx.dx>, <dx.dy>, <dy,dy>, <dx.dz>, <dy.dz>, <dz.dz> };
		//
		//              +-          -+           +-   -+
		//         xv = | r*cos(phi) |      mv = | r   |
		//              | r*sin(phi) |           | phi |
		//              | z          |           | z   |
		//              +-          -+           +-   -+
		//         
		//              +-                   -+    +-                        -+
		//              | @x/@r @x/@phi @x/@z |    | cos(phi)  -r*sin(phi)  0 |
		//    @xv/@mv = | @y/@r @y/@phi @y/@z |  = | sin(phi)   r*cos(phi)  0 |
		//              | @z/@r @z/@phi @z/@z |    |    0           0       1 |
		//              +-                   -+    +-                        -+
		//
		//              +-                               -+   +-                      -+
		//              | <dr.dr>   <dr.dphi>   <dr.dz>   |   |  0      0          0   |
		//       C_mv = | <dphi.dr> <dphi.dphi> <dphi.dz> | = |  0   (drphi/r)^2   0   |
		//              | <dz.dr>   <dz.dphi>   <dz.dz>   |   |  0      0         dz^2 |
		//              +-                               -+   +-                      -+
		//
		//     covMat = (@x/@mv) C_mv (@x/@mv)^T
		//              +-                             -+ +-                          -+
		//              | 0  -sin(phi)*(drphi^2)/r   0  | |    cos(phi)  sin(phi)    0 | 
		//            = | 0   cos(phi)*(drphi^2)/r   0  | | -r*sin(phi)  r*cos(phi)  0 |
		//              | 0           0            dz^2 | |    0             0       1 |
		//              +-                             -+ +-                          -+
		//              +-                                                               -+ 
		//              |  (drphi)^2*sin^2(phi)         -(drphi)^2*sin(phi)*cos(phi)  0   |
		//            = | -(drphi)^2*sin(phi)*cos(phi)   (drphi)^2*cos^2(phi)         0   |
		//              |           0                                0               dz^2 |
		//              +-                                                               -+
		//--
		float covMatrix  [] = { 0.F, 0.F, 0.F, 0.F, 0.F, 0.F };
		float drphi2 = pow(drphi, 2);
		float sinphi = sin(phi);
		float cosphi = cos(phi);
		covMatrix[0] =   drphi2 * sinphi * sinphi;
		covMatrix[1] = - drphi2 * sinphi * cosphi;
		covMatrix[2] = - drphi2 * cosphi * cosphi;
		covMatrix[5] =   dz*dz;

                //  Type of raw data hit, either one of
                //       LCIO::TPCHIT
                //       LCIO::SIMTRACKERHIT
	        newHitp->setType(hitp->getType());
		
		newHitp->setPosition(newPosition);
		newHitp->setCovMatrix(covMatrix); // FIXME: covMatrix is in local frame now

		newHitp->setEDep(_chargeConversionFactor * charge);
	        newHitp->setEDepError(_chargeConversionFactor * dcharge);

		if (pulsep) newHitp->setTime(pulsep->getTime());
		else        newHitp->setTime(static_cast<TrackerPulse *>(hitp->getRawHits()[0])->getTime());

		// FIXME: now simply copy all the pulses in the original hit
		auto pulsesInHit = hitp->getRawHits(); // vector containing TrackerPulses*
		for (auto &lcop : pulsesInHit) { // loop over pulses stored as LCObjects
			auto pulp = static_cast<TrackerPulse *>(lcop);
			newHitp->rawHits().push_back(pulp);
		}

		auto qualityFlag  = hitp->getQuality();
		newHitp->setQuality(qualityFlag);

		return newHitp;
	}

} // close namespace marlintpc
