//#define KF_DEBUG
#define KF_RETRY
//#define KF_EXP
#include "RootFileProcessor3.h"

// STL
#include <iostream>
#include <string>
#include <tuple>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

// GEAR
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/gearimpl/TPCModuleImpl.h"
#include "gearxml/GearXML.h"

// KalTest
#include "kaltest/TKalTrackSite.h"

// KalDetecotr, MeasLayer and KalHit
#include "kaldet/EXTPCKalDetector.h"
#include "kaldet/EXTPCMeasLayer.h"
#include "kaldet/EXTPCHit.h"
#include "EXHYBTrack.h"

// global constants from Marlin
#include "marlin/Global.h"

// exception handlers
#include "marlin/Exceptions.h"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// ROOT
#include "TMath.h"
#include "TVector3.h"
#include "TObjArray.h"

#ifdef MARLIN_USE_AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>
#endif
// ----- include for verbosity dependend logging ---------
#include "marlin/VerbosityLevels.h"


using namespace lcio;
using namespace marlin;
using namespace marlintpc;

// FIXME: to be moved to more appropriate place ///
const unsigned int kDiscarded = 1 << 20;
///////////////////////////////////////////////////

const int NMAXROWS = 84; // LP TPC case

#if 1
static const bool gkDir = kIterBackward ;
#else
static const bool gkDir = kIterForward;
#endif
static EXTPCKalDetector* _kaldetPtr = 0 ;

RootFileProcessor3 aRootFileProcessor3 ;


RootFileProcessor3::RootFileProcessor3() : Processor("RootFileProcessor3")
{
	  // modify processor description
	  _description = "RootFileProcessor3 makes ROOT file for TPC data analysis";
	  
	
	  // register steering parameters: name, description, class-variable, default value
	
	  registerInputCollection(LCIO::TRACK,
				  "InputCollectionName", 
				  "Name of the refined TPCTrack collection",
				  _inputColName,
				  std::string("RefinedTPCTracks"));
	
	  registerInputCollection(LCIO::TRACK,
				  "CurTrackCollectionName", 
				  "Name of the TPC Track collection",
				  _curTrackColName,
				  std::string("TPCTracks"));
	
	  registerInputCollection(LCIO::TRACKERRAWDATA,
				  "MergedRawDataCollectionName", 
				  "Name of the merged TPC raw data collection",
				  _mergedRawDataColName,
				  std::string("MergedTPCRawData"));
	
	  registerOptionalParameter("ModulesToUseOverride",
	  			    "Optional: modules to use",
	  			    _modulesToUseOverride,
	  			    _modulesToUseOverride);
	  
	  registerOptionalParameter("RowsToSkipOverride",
				    "Optional: rows to skip",
				    _badRowsOverride,
				    _badRowsOverride);
}


void RootFileProcessor3::init()
{ 
  	streamlog_out(DEBUG) << "   init called  " << std::endl << std::flush;

  	_nRun       = 0 ;
  	_nEvt       = 0 ;
  	_nMergedEvt = 0 ;

  	//-- 
  	// Set Ntuple used for TPC analysis
  	//-- 
  	setNtuple() ;

  	//-- 
  	// Usually a good idea to
  	//-- 
  	printParameters() ;

  	//-- 
  	// Create kaldet
  	//-- 
  	_kaldetPtr = EXTPCKalDetector::GetInstance() ;
  	
  	//--
  	// Get module geometry information from gear.
  	//--
  	const auto &theTPCParameters = marlin::Global::GEAR->getTPCParameters();
  	const auto nMod = theTPCParameters.getNModules();      // # modules
  	for (auto i_mod = 0; i_mod < nMod; i_mod++) _moduleVec.push_back(&theTPCParameters.getModule(i_mod));

	_maxDriftLength = theTPCParameters.getMaxDriftLength();
}


void RootFileProcessor3::processRunHeader([[maybe_unused]] LCRunHeader* run) 
{ 
  	_nRun++ ;
} 


void RootFileProcessor3::processEvent(LCEvent * evt)
{ 
	static auto tobuffer      = true;
	static auto nomergedtrack = true;
	static auto bufferempty   = true;

	static std::map<int, TrackInfoMap_t> *trackInfoMapVecPtr = nullptr;
	static std::map<int, HitInfoMap_t  > *hitInfoMapVecPtr   = nullptr;
	static std::map<int, PadInfoMap_t  > *padInfoMapVecPtr   = nullptr;

	//--
	// Begin event message.
	//--
	_nEvt = evt->getEventNumber();
        streamlog_out(DEBUG) << std::endl
                             << "***** RootFileProcessor3::processEvent(...) ********************************" << std::endl
                             << "   Event No: "    << _nEvt << "  Merged Event No: " << _nMergedEvt       
			     << "   tobuffer: "    << (tobuffer    ? "true" : "false")
			     << "   bufferempty: " << (bufferempty ? "true" : "false")		               << std::endl
                             << "****************************************************************************" << std::endl
                             << std::flush;
	//--
	// Make sure "merged raw data collection" there:
	//   1) if not  : the event was not used to create a pseudo 2-track event.
	//   2) if there: the event was used to create a pseudo 2-track event.
	//   	 2-1) if empty   : it's 1st event got merged (tobuffer = true)
	//   	 2-2) if nonempty: it's 2nd event got merged (tobuffer = false)
	//--
	auto nRawData = 0;
  	EVENT::LCCollection* colMergedRawDataPtr = nullptr;
  	try {
  	  	colMergedRawDataPtr = evt->getCollection(_mergedRawDataColName);
		nRawData  = colMergedRawDataPtr->getNumberOfElements();
		tobuffer  = nRawData == 0; // save info to buffer if true
  	} catch (std::exception &) {
  	  	streamlog_out( WARNING ) << "This event has no merged raw data collection! Skipped." << std::endl;
		return;
  	} 

  	//--
  	// Get input collections.
  	//--
	_nKalTrks = 0;
  	EVENT::LCCollection* mergedColTracksPtr = nullptr;
	auto nMergedTrks =  0;
  	try {
  	  	mergedColTracksPtr = evt->getCollection(_inputColName);
	  	nMergedTrks = mergedColTracksPtr->getNumberOfElements();
		nomergedtrack = nMergedTrks == 0;

  	} catch (std::exception &) {
		if (!tobuffer) {
			streamlog_out( WARNING ) << "This event has no merged Track collection!" << std::endl;
		}
		nomergedtrack = true;
  	} 

  	EVENT::LCCollection* curColTracksPtr = nullptr;
  	try {
  	  	curColTracksPtr = evt->getCollection(_curTrackColName);
  	} catch (std::exception &) {
  	  	streamlog_out( WARNING ) << "This event has no Track collection! Skipped." << std::endl;
		//--
		// Clean up, before skipping this event that was used to create a merged event
		//--
		delete trackInfoMapVecPtr; trackInfoMapVecPtr = nullptr;
		delete hitInfoMapVecPtr  ; hitInfoMapVecPtr   = nullptr;
		delete padInfoMapVecPtr  ; padInfoMapVecPtr   = nullptr;

		tobuffer      = true;
		nomergedtrack = true;
		bufferempty   = true;

  	  	return;
  	} 

  	//--
  	// Now analyze this event together with the previous one, if any.
  	//--

	if (tobuffer) { // first event merged
		//--
		// Create buffers: index = (0, 1, 2) = (1st event, 2nd event, merged event).
		//--
		trackInfoMapVecPtr = new std::map<int, TrackInfoMap_t>;
		hitInfoMapVecPtr   = new std::map<int, HitInfoMap_t  >;
		padInfoMapVecPtr   = new std::map<int, PadInfoMap_t  >;

		auto &trackInfoMapVec = *trackInfoMapVecPtr;
		auto &hitInfoMapVec   = *hitInfoMapVecPtr;
		auto &padInfoMapVec   = *padInfoMapVecPtr;
		
  		//--
  		// Fill info maps for this event.
  		//--
		getInfoMaps(curColTracksPtr, trackInfoMapVec[0], hitInfoMapVec[0], padInfoMapVec[0]);

		tobuffer     = false;
		bufferempty  = false;

		//--
		// Fill n-tuples for this event.
		//--
  		auto nTrks = curColTracksPtr->getNumberOfElements();
		fillTuples(nTrks, *_trks, *_hits, *_clrs, trackInfoMapVec[0], hitInfoMapVec[0], padInfoMapVec[0]);

	} else { // second event merged

		if (bufferempty) {
  	  		streamlog_out( ERROR ) << "Buffer not filled! Skip this merged pair of events" << std::endl;
			//--
			// Clean up the buffers.
			//--
			delete trackInfoMapVecPtr; trackInfoMapVecPtr = nullptr;
			delete hitInfoMapVecPtr  ; hitInfoMapVecPtr   = nullptr;
			delete padInfoMapVecPtr  ; padInfoMapVecPtr   = nullptr;

			tobuffer      = true;
			nomergedtrack = true;
			bufferempty   = true;

			return;
		}

		auto &trackInfoMapVec = *trackInfoMapVecPtr;
		auto &hitInfoMapVec   = *hitInfoMapVecPtr;
		auto &padInfoMapVec   = *padInfoMapVecPtr;

  		//--
  		// Fill info maps for this event.
  		//--
		getInfoMaps(curColTracksPtr, trackInfoMapVec[1], hitInfoMapVec[1], padInfoMapVec[1]);

		//--
		// Fill n-tuples for this event.
		//--
  		auto nTrks = curColTracksPtr->getNumberOfElements();
		fillTuples(nTrks, *_trks, *_hits, *_clrs, trackInfoMapVec[1], hitInfoMapVec[1], padInfoMapVec[1]);
		
		//--
		// Initialize the merged event maps with default values.
		//     (serialrow, nhits=0, ntrks=0, evtno, mevtno, rphic1st, zc1st, rphi2nd, zc2nd)
		// assuming that both 1st event and 2nd event have one and only one track per event, which
		// should be assured by EventMergerProcessor.
		//--
		initInfoMaps(hitInfoMapVec[0], hitInfoMapVec[1], trackInfoMapVec[2], hitInfoMapVec[2], padInfoMapVec[2]);

  		//--
  		// Fill info maps for the merged event.
  		//--
		if (!nomergedtrack) { // found tracks in the merged event

			auto ismerged = true;
			getInfoMaps(mergedColTracksPtr, trackInfoMapVec[2], hitInfoMapVec[2], padInfoMapVec[2], ismerged);
		}

  		//--
  		// Fill n-tuples for merged events.
  		//--
		fillTuples(nMergedTrks, *_trks2, *_hits2, *_clrs2, trackInfoMapVec[2], hitInfoMapVec[2], padInfoMapVec[2]);

  	  	streamlog_out( DEBUG ) << "     nMergedTrks = " << nMergedTrks << std::endl;

		//--
		// Clean up the buffers.
		//--
		delete trackInfoMapVecPtr; trackInfoMapVecPtr = nullptr;
		delete hitInfoMapVecPtr  ; hitInfoMapVecPtr   = nullptr;
		delete padInfoMapVecPtr  ; padInfoMapVecPtr   = nullptr;

		tobuffer      = true;
		nomergedtrack = true;
		bufferempty   = true;

		//--
		// Count # merged events.
		//--
  		_nMergedEvt++;
	}
}


void RootFileProcessor3::check([[maybe_unused]] LCEvent * evt)
{ 
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void RootFileProcessor3::end()
{ 
        streamlog_out( DEBUG4 )  << std::endl  
                                 << "***** RootFileProcessor3::end() ******" << std::endl << std::flush;
    
        streamlog_out( MESSAGE ) << " ------------------------------------------------------- " << std::endl
    			         << "   " << _nMergedEvt 
    			         << " merged events analyzed in " << _nEvt << " input events."  << std::endl
    			         << std::flush;
}

#ifdef MARLIN_USE_AIDA
void RootFileProcessor3::setNtuple()
{
	  /** For storing cluster data 
	   */
	  std::vector<std::string> _clrsItemNames;
	  std::vector<std::string> _clrsItemType;
	
	  /** For storing hit data 
	   */
	  std::vector<std::string> _hitsItemNames;
	  std::vector<std::string> _hitsItemType;
	
	  /** For storing track data 
	   */
	  std::vector<std::string> _trksItemNames;
	  std::vector<std::string> _trksItemType;
	
	  // cluster info
	  _clrsItemNames.push_back("module");
	  _clrsItemType.push_back("int");
	
	  _clrsItemNames.push_back("row");
	  _clrsItemType.push_back("int");
	
	  _clrsItemNames.push_back("normchg");
	  _clrsItemType.push_back("float");
	
	  _clrsItemNames.push_back("drphi");
	  _clrsItemType.push_back("float");
	
	  _clrsItemNames.push_back("dt");
	  _clrsItemType.push_back("float");
	
	  _clrsItemNames.push_back("chgsum");
	  _clrsItemType.push_back("float");
	
	  _clrsItemNames.push_back("ndf");
	  _clrsItemType.push_back("int");
	
	  _clrsItemNames.push_back("chi2");
	  _clrsItemType.push_back("float");
	
	  _clrsItemNames.push_back("fi0");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("cpa");
	  _clrsItemType.push_back("float");
	
	  _clrsItemNames.push_back("tnl");
	  _clrsItemType.push_back("float");
	
	  _clrsItemNames.push_back("ntrks");
	  _clrsItemType.push_back("int");
	
	  _clrsItemNames.push_back("fi0loc");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("fiinc");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("trkid");
	  _clrsItemType.push_back("int");
	
	  _clrsItemNames.push_back("evtno");
	  _clrsItemType.push_back("int");
	
	  _clrsItemNames.push_back("rphic");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("zc");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("mevtno");
	  _clrsItemType.push_back("int");
	
	  _clrsItemNames.push_back("rphixp1");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("zxp1");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("rphixp2");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("zxp2");
	  _clrsItemType.push_back("double");
	
	  _clrsItemNames.push_back("nktrks");
	  _clrsItemType.push_back("int");
	

	  // hit info
	  _hitsItemNames.push_back("serialrow");
	  _hitsItemType.push_back("int");
	
	  _hitsItemNames.push_back("dxin");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("dxot");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("dzin");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("dzot");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("chg");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("ndf");
	  _hitsItemType.push_back("int");
	
	  _hitsItemNames.push_back("chi2");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("fi0");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("cpa");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("tnl");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("nhits");
	  _hitsItemType.push_back("int");
	
	  _hitsItemNames.push_back("ntrks");
	  _hitsItemType.push_back("int");
	
	  _hitsItemNames.push_back("fi0loc");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("fiinc");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("rphic");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("trkid");
	  _hitsItemType.push_back("int");
	
	  _hitsItemNames.push_back("evtno");
	  _hitsItemType.push_back("int");
	
	  _hitsItemNames.push_back("zc");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("mevtno");
	  _hitsItemType.push_back("int");
	
	  _hitsItemNames.push_back("rphixp1");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("zxp1");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("rphixp2");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("zxp2");
	  _hitsItemType.push_back("double");
	
	  _hitsItemNames.push_back("nktrks");
	  _hitsItemType.push_back("int");

	
	  // track info
	  _trksItemNames.push_back("ndf");
	  _trksItemType.push_back("int");
	
	  _trksItemNames.push_back("chi2");
	  _trksItemType.push_back("double");
	
	  _trksItemNames.push_back("fi0");
	  _trksItemType.push_back("double");
	
	  _trksItemNames.push_back("cpa");
	  _trksItemType.push_back("double");
	
	  _trksItemNames.push_back("tnl");
	  _trksItemType.push_back("double");
	
	  _trksItemNames.push_back("ntrks");
	  _trksItemType.push_back("int");
	
	  _trksItemNames.push_back("trkid");
	  _trksItemType.push_back("int");
	
	  _trksItemNames.push_back("evtno");
	  _trksItemType.push_back("int");
	
	  _trksItemNames.push_back("mevtno");
	  _trksItemType.push_back("int");
	
	  //--
	  // Create tuples.
	  //--
	  // tuple paths and names
	  _tuplePathTitleVec["unmerged"]["trks"] = { "trks", "trks" };
	  _tuplePathTitleVec["unmerged"]["hits"] = { "hits", "hits" };
	  _tuplePathTitleVec["unmerged"]["clrs"] = { "clrs", "clrs" };

	  _tuplePathTitleVec["merged"]["trks"] = { "trks2", "trks2" };
	  _tuplePathTitleVec["merged"]["hits"] = { "hits2", "hits2" };
	  _tuplePathTitleVec["merged"]["clrs"] = { "clrs2", "clrs2" };

	  // for unmerged events
	  auto [clrsPath, clrsTitle] = _tuplePathTitleVec["unmerged"]["clrs"];
	  _clrs = AIDAProcessor::tupleFactory(this)
		    ->create(clrsPath, clrsTitle, _clrsItemNames, _clrsItemType, "");
	
	  auto [hitsPath, hitsTitle] = _tuplePathTitleVec["unmerged"]["hits"];
	  _hits = AIDAProcessor::tupleFactory(this)
		    ->create(hitsPath, hitsTitle, _hitsItemNames, _hitsItemType, "");
	
	  auto [trksPath, trksTitle] = _tuplePathTitleVec["unmerged"]["trks"];
	  _trks = AIDAProcessor::tupleFactory(this)
		    ->create(trksPath, trksTitle, _trksItemNames, _trksItemType, "");

	  // for unmerged events
	  auto [clrsPath2, clrsTitle2] = _tuplePathTitleVec["merged"]["clrs"];
	  _clrs2 = AIDAProcessor::tupleFactory(this)
		    ->create(clrsPath2, clrsTitle2, _clrsItemNames, _clrsItemType, "");
	
	  auto [hitsPath2, hitsTitle2] = _tuplePathTitleVec["merged"]["hits"];
	  _hits2 = AIDAProcessor::tupleFactory(this)
		    ->create(hitsPath2, hitsTitle2, _hitsItemNames, _hitsItemType, "");
	
	  auto [trksPath2, trksTitle2] = _tuplePathTitleVec["merged"]["trks"];
	  _trks2 = AIDAProcessor::tupleFactory(this)
		    ->create(trksPath2, trksTitle2, _trksItemNames, _trksItemType, "");
}
#endif

using namespace std;

/// utility functions
int RootFileProcessor3::getModuleID(const TrackerHit &hit)
{
    	TrackerPulse* plsp = static_cast<TrackerPulse *>(hit.getRawHits()[0]);
    	int mod = plsp->getCellID1();

    	return mod;
}

int RootFileProcessor3::getRowID(const TrackerHit &hit , const gear::TPCModule &tpcmodule)
{
  	auto plsp  = static_cast<TrackerPulse *>(hit.getRawHits()[0]);
  	auto padID = plsp->getCellID0();
  	auto rowID = tpcmodule.getRowNumber(padID);

  	return rowID ;
}

int RootFileProcessor3::getSerialRowNo(int modID, int rowID)
{
 	auto  superlayer = (modID < 2 ? 0 : (modID < 5 ? 1 : 2));
 	auto  rowSerial  = superlayer * _moduleVec[modID]->getNRows() + rowID; 

	return rowSerial;
}

bool RootFileProcessor3::isMultiHit(const EXTPCHit &kalhit)
{
	auto &hit = *static_cast<const TrackerHit *>(kalhit.GetHitPtr());
	auto qualityFlag = hit.getQuality();

	return hitflag::isMultipleHitCandidate(qualityFlag);
}

bool RootFileProcessor3::isSplitHit(const EXTPCHit &kalhit)
{
	auto &hit = *static_cast<const TrackerHit *>(kalhit.GetHitPtr());
	auto qualityFlag = hit.getQuality();
	LCFlagImpl aflag(qualityFlag);
	// FIXME: //////////////////////////////////////////////////////////
	//    to be moved to more appropriate place
	int const SPLIT_HIT_BIT_INDEX = 20; // bit 20
	////////////////////////////////////////////////////////////////////

	return aflag.bitSet(SPLIT_HIT_BIT_INDEX);
}


/// initialize event information map
void RootFileProcessor3::initInfoMaps(HitInfoMap_t   & hitInfoMap1,  // hit info for 1st event (input)
				      HitInfoMap_t   & hitInfoMap2,  // hit info for 2nd event (input)
		     [[maybe_unused]] TrackInfoMap_t & trackInfoMap, // track info (output)
				      HitInfoMap_t   & hitInfoMap,   // hit   info (output)
		     [[maybe_unused]] PadInfoMap_t   & padInfoMap)   // pad   info (output)
{ 
	//--
	// Make sure that both of this merged event have exactly one track each before merging.
	//--
	auto nTrks1 = std::get<EHntrks>(hitInfoMap1[0][0]);
	auto nTrks2 = std::get<EHntrks>(hitInfoMap2[0][0]);
	if (nTrks1 != 1 || nTrks2 != 1) {
                StopProcessingException skip(this);
                streamlog_out( ERROR ) << "The number of good TPC tracks in at least one of unmerged events was not 1!"
					  "     nTrks1: " << nTrks1 << "  nTrks2: " <<  nTrks2 << "!  Skipped."
                                       << std::endl << std::flush;
		return;
	}

	//--
	// Get expected hit positions on each row from hitInfoMaps of unmerged events.
	//--
	std::vector<std::tuple<double, double, double, double>> expectedRphiZ;
  	for (int irow = 0; irow < NMAXROWS; irow++) { 
		auto rphixp1  = std::get<EHrphic>(hitInfoMap1[0][irow]);
		auto zxp1     = std::get<EHzc>   (hitInfoMap1[0][irow]);
		auto rphixp2  = std::get<EHrphic>(hitInfoMap2[0][irow]);
		auto zxp2     = std::get<EHzc>   (hitInfoMap2[0][irow]);
		expectedRphiZ.push_back({rphixp1, zxp1, rphixp2, zxp2});
	}

  	//--
  	// First initialize the hitInfoMap.
  	//--
	// FIXME: /////////////////////////////////////////////////
  	auto nTrksMax = 10; // allow up to 10 tracks in merged event
	///////////////////////////////////////////////////////////
  	for (auto i_trk = 0; i_trk < nTrksMax; i_trk++) { // track loop starts here
  	  	for (int irow = 0; irow < NMAXROWS; irow++) { 
  			//--
  			// First initialize the hitInfoMap.
  			//--
			auto dxin   = -1.e6;
			auto dxot   = -1.e6;
			auto dzin   = -1.e6;
			auto dzot   = -1.e6;
			auto chg    = -1.e6;
			auto ndf    =     0;
			auto chi2   =  1.e9;
			auto fi0    = -1.e6;
			auto cpa    = -1.e6;
			auto tnl    = -1.e6;
			auto nhits  =     0;
			auto ntrks  =     0;
			auto fi0loc = -1.e6;
			auto fiinc  = -1.e6;
			auto rphic  = -1.e6;
			auto trkid  = i_trk;
			auto evtno  = _nEvt;
			auto zc     = -1.e6;
			auto mevtno = _nMergedEvt;
			auto [rphixp1, zxp1, rphixp2, zxp2] = expectedRphiZ[irow];

			hitInfoMap[i_trk][irow] = {dxin , dxot   , dzin  , dzot  , chg    ,
						   ndf  , chi2   , fi0   , cpa   , tnl    ,
						   nhits, ntrks  , fi0loc, fiinc , rphic  ,
						   trkid, evtno  , zc    , mevtno, rphixp1, 
						   zxp1 , rphixp2, zxp2  };
    		} // end of row loop
  	} //end of track loop
}


/// Get event information map
void RootFileProcessor3::getInfoMaps(LCCollection   * colTracksPtr,
				     TrackInfoMap_t & trackInfoMap,
				     HitInfoMap_t   & hitInfoMap,
				     PadInfoMap_t   & padInfoMap,
				     bool             ismerged)     // true if this is a pseudo 2-track event
{ 
  	//--
  	// Get TPC geometry info.
  	//--
  	const auto &theTPCParameters = marlin::Global::GEAR->getTPCParameters();

  	//--
  	// Loop over tracks.
  	//--
	TObjArray kaltracks;
	kaltracks.SetOwner();

  	auto nTrks = colTracksPtr->getNumberOfElements(); // # tracks
  	for (auto i_trk = 0; i_trk < nTrks; i_trk++) { // track loop starts here
  	  	auto trkp = static_cast<IMPL::TrackImpl *>(colTracksPtr->getElementAt(i_trk));
		auto &trk = *trkp;

  	  	//--
  	  	// Count hits per row and calculate (rphi, z) of each hit.
  	  	//--
		// initialize hit counter
		std::map<int, int> nHits;
  	  	for (int irow = 0; irow < NMAXROWS; irow++) nHits[irow] = 0;

		// initialize row hit position to expected one
		std::map<int, std::tuple<double, double>> rphizhit;
  	  	for (int irow = 0; irow < NMAXROWS; irow++) {
			auto  xvexpect  = getExpectedHitPosition(trk, irow); // expected hit position in this row
			if (abs(xvexpect.Z()) < 1.e4) {
				auto &tpcmodule = theTPCParameters.getNearestModule(xvexpect.X(), xvexpect.Y(), xvexpect.Z());
				auto rphiv      = tpcmodule.globalToLocal(xvexpect.X(), xvexpect.Y()); // in module-local frame (polar)

				rphizhit[irow]  = {rphiv[0]*rphiv[1], xvexpect.Z()};
			} else {
				rphizhit[irow]  = { -1.e6, -1.e6 }; // no crossing point found
			}
		}

		// initialize row to hit map
		std::map<int, EVENT::TrackerHit *> rowToHitMap;
  	  	for (int irow = 0; irow < NMAXROWS; irow++) {
			rowToHitMap[irow] = nullptr;
		}

		// fill the hit info
  	  	const EVENT::TrackerHitVec &hitvec = trk.getTrackerHits() ;
  	  	for (auto objp : hitvec) {
  	  	  	auto  hitp  = static_cast<EVENT::TrackerHit *>(objp);
			auto &hit   = *hitp;
  	  	  	auto  modID = getModuleID(hit); 

			// skip this hit if the module has been excluded by user
  	  	  	if (_modulesToUseOverride.end() == find(_modulesToUseOverride.begin(), _modulesToUseOverride.end(), modID)) continue;

			// fill row to hit map
			auto &tpcmodule  = *_moduleVec[modID];
  	  	  	auto  rowID      = getRowID(hit, tpcmodule);
  	  	  	auto  superlayer = (modID < 2 ? 0 : (modID < 5 ? 1 : 2)); // this works only for LP1
  	  	  	auto  rowSerial  = superlayer * tpcmodule.getNRows() + rowID; 
			rowToHitMap[rowSerial] = hitp;

			// count hits on the row
  	  	  	nHits[rowSerial]++;

			// set (rphi, z) coordinates for this hit
			auto hpos       = hit.getPosition();
			auto rphiv      = tpcmodule.globalToLocal(hpos[0], hpos[1]); // in module-local frame (polar)
			rphizhit[rowID] = {rphiv[0]*rphiv[1], hpos[2]}; 	     // update this info if real hit there
  	  	}

		// initialize hitInfoMap with unphysical default values
		// this is needed in case Kalman filter fails
  	  	for (int irow = 0; irow < NMAXROWS; irow++) { 
			auto dxin   = -1.e6;
			auto dxot   = -1.e6;
			auto dzin   = -1.e6;
			auto dzot   = -1.e6;
			auto chg    = -1.e6;
			auto ndf    =     0;
			auto chi2   =  1.e9;
			auto fi0    = -1.e6;
			auto cpa    = -1.e6;
			auto tnl    = -1.e6;
			auto fi0loc = -1.e6;
			auto fiinc  = -1.e6;
			auto mevtno = _nMergedEvt;

			auto [rphic, zc] = rphizhit[irow]; // (real/expected) hit position
			auto rphixp1     = -1.e6;
			auto zxp1        = -1.e6;
			auto rphixp2     = -1.e6;
			auto zxp2        = -1.e6;

			// restore expected hit positions if already set
			if (ismerged) {
				rphixp1  = std::get<EHrphixp1>(hitInfoMap[i_trk][irow]);
				zxp1     = std::get<EHzxp1>   (hitInfoMap[i_trk][irow]);
				rphixp2  = std::get<EHrphixp2>(hitInfoMap[i_trk][irow]);
				zxp2     = std::get<EHzxp2>   (hitInfoMap[i_trk][irow]);
			}

			hitInfoMap[i_trk][irow] = {dxin       , dxot   , dzin  , dzot  , chg    ,
						   ndf        , chi2   , fi0   , cpa   , tnl    ,
						   nHits[irow], nTrks  , fi0loc, fiinc , rphic  ,
						   i_trk      , _nEvt  , zc    , mevtno, rphixp1,
						   zxp1       , rphixp2, zxp2  };
  	  	}

  	  	//--
  	  	// Fit track again to apply additional cut and to get residual distribution 
  	  	//--
  	  	// create KalHits
  	  	TObjArray kalhits;
  	  	kalhits.SetOwner();

  	  	for (auto objp : hitvec) {
  	  	  	auto hitp  = static_cast<IMPL::TrackerHitImpl *>(objp);

			// skip this hit if the module has been excluded by user
  	  	  	auto  modID = getModuleID(*hitp); 
  	  	  	if (_modulesToUseOverride.end() == find(_modulesToUseOverride.begin(), _modulesToUseOverride.end(), modID)) continue;

  	  	  	kalhits.Add(createKalHit(hitp, ismerged));
  	  	} // end of hits loop

		// create KalTrack
#ifndef KF_RETRY
    		auto kaltrackp = createKalTrack(kalhits, gkDir, nHits);
#else
		EXHYBTrack *kaltrackp = nullptr;
		auto dir = gkDir;
		for (auto itry = 0; itry < 2; itry++) {
    			kaltrackp = createKalTrack(kalhits, dir, nHits);
			if (kaltrackp) {
				if (itry != 0) {
					streamlog_out(DEBUG) << " --> retry succeeded" << endl
							     << flush;
				}
				break;
			} else {
				if (itry == 0) {
					streamlog_out(DEBUG) << " Too many discarded hits, reverse dir and try again" << endl
							     << flush;
					dir = !dir;
				} else {
					streamlog_out(DEBUG) << " Too many discarded hits, Kalman retry failed" << endl
							     << flush;
				}
			}
		}
#endif
		if (!kaltrackp) continue;            // Kalman filter failed, skip this track

		kaltracks.Add(kaltrackp);
		auto &kaltrack = *kaltrackp;
      
                //--
		// Inverse Kalman filter this track to 1st site
		//--
    		kaltrack.SmoothBackTo(1);

                //--
		// Go to innermost hit
		//--
		auto isite1 = 1; // site 0 is dummy
#ifndef KF_RETRY
		if (gkDir == kIterBackward) isite1 = kaltrack.GetEntries() - 1;
#else
		if (dir == kIterBackward) isite1 = kaltrack.GetEntries() - 1;
#endif
		auto &firstSite = *static_cast<const TKalTrackSite *>(kaltrack.At(isite1));

                //--
		// Now fill n-tuples.
		//--
		// track information
    		auto ndf  = kaltrack.GetNDF() ;
    		auto chi2 = kaltrack.GetChi2() ;
    		auto fi0  = firstSite.GetCurState()(1, 0); 
    		auto cpa  = firstSite.GetCurState()(2, 0);
    		auto tnl  = firstSite.GetCurState()(4, 0);

		trackInfoMap[i_trk] = TrackInfo_t{ndf, chi2, fi0, cpa, tnl, nTrks, i_trk, _nEvt, _nMergedEvt};

		// hit and cluster information
    		TIter nextlayer(&kaltrack);
    		nextlayer();
    		TKalTrackSite* sitep;
    		while ((sitep = static_cast<TKalTrackSite *>(nextlayer()))) { // loop over sites
			// get KalHit and TrkerHit
			auto &site   = *sitep;
			auto &kalhit = static_cast<const EXTPCHit &>(site.GetHit());
    		  	auto  hitp   = static_cast<const TrackerHit *>(kalhit.GetHitPtr());
    		  	auto &hit    = *hitp;

			// calculate incident phi angle
			auto modID   = getModuleID(hit); 
			auto rowID   = getRowID(hit, *_moduleVec[modID]);
			auto [fiinc, xvexpect] = getIncidentPhiAndPosition(kaltrack, kalhit, *_moduleVec[modID]);
#if 0
			auto &measlyr = site.GetHit().GetMeasLayer(); 
      			auto  irow   = measlyr.GetIndex();             // works only for LP1
#else
			auto  irow   = getSerialRowNo(modID, rowID);
#endif
			// retrieve default hit information
			auto [dxind  , dxotd  , dzind  , dzotd  , chgd   ,
			      ndfd   , chi2d  , fi0d   , cpad   , tnld   ,
			      nhitsd , ntrksd , fi0locd, fiincd , rphicd ,
			      trkidd , evtnod , zcd    , mevtnod, rphixp1,
			      zxp1   , rphixp2, zxp2   ] = hitInfoMap[i_trk][irow];

			// calculate local position in (r,phi)
      			auto lpos    = _moduleVec[modID]->globalToLocal(xvexpect[0], xvexpect[1]); // local in polar
                        auto rc      = lpos[0];
			auto phic    = lpos[1];

			auto rphic   = rc * phic;
			auto zc      = xvexpect[2];

			// calculate local fi0
    		  	auto fi0loc  = site.GetCurState()(1, 0); 

			// now update the hit information
      			auto dxin = site.GetResVec()(0,0);
      			auto dzin = site.GetResVec()(1,0);
      			auto dxot = 1.e10;
      			auto dzot = 1.e10;
      			auto chg  = hit.getEDep();

      			site.InvFilter(); // inverse Kalman filter

      			dxot = sitep->GetResVec( TVKalSite::kInvFiltered)(0,0);
      			dzot = sitep->GetResVec( TVKalSite::kInvFiltered)(1,0);

			auto evtno  = _nEvt;
			auto mevtno = _nMergedEvt;

			hitInfoMap[i_trk][irow] = {dxin       , dxot   , dzin  , dzot  , chg    ,
						   ndf        , chi2   , fi0   , cpa   , tnl    ,
						   nHits[irow], nTrks  , fi0loc, fiinc , rphic  ,
						   i_trk      , evtno  , zc    , mevtno, rphixp1,
						   zxp1       , rphixp2, zxp2  };
     
			// now fill cluster information
      			auto &pulseVec = hit.getRawHits();
      			for (auto objp : pulseVec) { // loop over pulses
      				auto  plsp  = static_cast<const IMPL::TrackerPulseImpl *>(objp);
      			  	auto &pls   = *plsp;
      			  	auto  padID = pls.getCellID0();
      			  	auto  ppos  = _moduleVec[modID]->getPadCenter(padID); // gear::Vector2D (Cartesian)
      			  	auto  lppos = _moduleVec[modID]->globalToLocal(ppos[0], ppos[1]); // local in polar

      			  	auto  r   = lppos[0];
      			  	auto  phi = lppos[1];
      			  	auto  dt  = pls.getTime();

      			  	auto hitchg  = hit.getEDep() ;
      			  	auto normchg = pls.getCharge()/hitchg ;
      			  	auto drphi   = r * phi - rc * phic ;

				padInfoMap[i_trk][irow].push_back({modID , rowID  , normchg, drphi , dt     ,
								   hitchg, ndf    , chi2   , fi0   , cpa    ,
								   tnl   , nTrks  , fi0loc , fiinc , i_trk  ,
								   evtno , rphic  , zc     , mevtno, rphixp1,
								   zxp1  , rphixp2, zxp2   });
      			} // end of pulse loop
    		} // end of site loop
  	} //end of track loop
	_nKalTrks = kaltracks.GetEntries();
}


/// fill track info ntuple
void RootFileProcessor3::fillTupTrk(const int      i_trk,     // track #
			            AIDA::ITuple & trks,      // tuple to fill
			            TrackInfo_t  & trackInfo) // track info
{
	// fill track data
	auto [ndf, chi2, fi0, cpa, tnl, ntrks, trkid, evtno, mevtno] = trackInfo;

    	trks.fill(0, ndf   );
    	trks.fill(1, chi2  );
    	trks.fill(2, fi0   );
    	trks.fill(3, cpa   );
    	trks.fill(4, tnl   );
    	trks.fill(5, ntrks );
    	trks.fill(6, i_trk );
    	trks.fill(7, evtno );
    	trks.fill(8, mevtno);

    	trks.addRow();
}	


/// fill hit info ntuple
void RootFileProcessor3::fillTupHit(const int      i_trk,   // track #
				    const int      irow,    // serial row #
			            AIDA::ITuple & hits,    // tuple to fill
			            HitInfo_t    & hitInfo) // hit info
{
	auto [dxin , dxot   , dzin  , dzot  , chg    ,
	      ndf  , chi2   , fi0   , cpa   , tnl    ,
	      nhits, ntrks  , fi0loc, fiinc , rphic  ,
	      trkid, evtno  , zc    , mevtno, rphixp1,
	      zxp1 , rphixp2, zxp2  ] = hitInfo;

      	hits.fill( 0, irow   );
      	hits.fill( 1, dxin   );
      	hits.fill( 2, dxot   );
      	hits.fill( 3, dzin   );
      	hits.fill( 4, dzot   );
      	hits.fill( 5, chg    );
      	hits.fill( 6, ndf    );
      	hits.fill( 7, chi2   );
      	hits.fill( 8, fi0    );
      	hits.fill( 9, cpa    );
      	hits.fill(10, tnl    );
      	hits.fill(11, nhits  );
      	hits.fill(12, ntrks  );
      	hits.fill(13, fi0loc );
      	hits.fill(14, fiinc  );
      	hits.fill(15, rphic  );
      	hits.fill(16, i_trk  ); // i_trk < 0 if imaginary track
	hits.fill(17, evtno  );
	hits.fill(18, zc     );
	hits.fill(19, mevtno );
	hits.fill(20, rphixp1);
	hits.fill(21, zxp1   );
	hits.fill(22, rphixp2);
	hits.fill(23, zxp2   );
	hits.fill(24, _nKalTrks);
	
	hits.addRow();
#ifdef KF_DEBUG
	if (irow > -1 && irow < 9999) {
		streamlog_out( DEBUG ) << "      irow = " << irow
				       << "  i_trk = "    << i_trk
				       << "  trkid = "    << trkid
				       << "  nhits = "    << nhits
				       << "  ntrks = "    << ntrks
				       << "  rphixp1 = "  << rphixp1
				       << "  zxp1 = "     << zxp1
				       << "  rphixp2 = "  << rphixp2
				       << "  zxp2 = "     << zxp2
				       << std::endl << std::flush;
	}
#endif
}


/// fill pad info ntuple
void RootFileProcessor3::fillTupClr(const int                i_trk,      // track #
			            AIDA::ITuple           & clrs,       // tuple to fill
			            std::vector<PadInfo_t> & padInfoVec) // pad info vector
{
	for (auto padInfo : padInfoVec) {
		auto [modid , row    , normchg, drphi , dt     ,
		      hitchg, ndf    , chi2   , fi0   , cpa    ,
		      tnl   , ntrks  , fi0loc , fiinc , trkid  ,
		      evtno , rphic  , zc     , mevtno, rphixp1,
		      zxp1  , rphixp2, zxp2   ] = padInfo;

      		clrs.fill( 0, modid  );
      		clrs.fill( 1, row    );
      		clrs.fill( 2, normchg);
      		clrs.fill( 3, drphi  );
      		clrs.fill( 4, dt     );
      		clrs.fill( 5, hitchg );
      		clrs.fill( 6, ndf    );
      		clrs.fill( 7, chi2   );
      		clrs.fill( 8, fi0    );
      		clrs.fill( 9, cpa    );
      		clrs.fill(10, tnl    );
      		clrs.fill(11, ntrks  );
      		clrs.fill(12, fi0loc );
      		clrs.fill(13, fiinc  );
      		clrs.fill(14, i_trk  );
      		clrs.fill(15, evtno  );
      		clrs.fill(16, rphic  );
      		clrs.fill(17, zc     );
      		clrs.fill(18, mevtno );
      		clrs.fill(19, rphixp1);
      		clrs.fill(20, zxp1   );
      		clrs.fill(21, rphixp2);
      		clrs.fill(22, zxp2   );
      		clrs.fill(23, _nKalTrks);

      		clrs.addRow();
	}
}


/// Fill output n-tuples
void RootFileProcessor3::fillTuples(const int        nTrks,
				    AIDA::ITuple   & trks,
				    AIDA::ITuple   & hits,
				    AIDA::ITuple   & clrs,
				    TrackInfoMap_t & trackInfoMap,
				    HitInfoMap_t   & hitInfoMap,
				    PadInfoMap_t   & padInfoMap)
{
	streamlog_out( DEBUG ) << "-----------------------------------" << std::endl
			       << " fillTuples: "                       << std::endl
			       << "     trks: title = " << trks.title() << std::endl
			       << "           nTrks = " << nTrks        << std::endl
			       << "-----------------------------------" << std::endl
			       << std::flush;
  	//--
  	// Check n-tuple type
  	//--
	  	auto [trksPath2, trksTitle2] = _tuplePathTitleVec["merged"]["trks"];
		if (!nTrks && trks.title() == trksTitle2) {

  	  		streamlog_out( WARNING ) << "fillTuples: nTrks = 0 for this merged event" 
					         << std::endl << std::flush;

			// fill an imaginary track with no hits but with expected hits info from unmerged events
			auto i_trk = -1; // imaginary track
  	  		for (int irow = 0; irow < NMAXROWS; irow++) {  // loop over rows
				auto &hitInfo = hitInfoMap[0][irow];
				fillTupHit(i_trk, irow, hits, hitInfo);
			}				
		}

  	//--
  	// Now fill n-tuples to output.
  	//--
  	for (auto i_trk = 0; i_trk < nTrks; i_trk++) { // track loop starts here

		// fill track data
		fillTupTrk(i_trk, trks, trackInfoMap[i_trk]);

		// fill hit data
  	  	for (int irow = 0; irow < NMAXROWS; irow++) {  // loop over rows
			try {
				auto &hitInfo = hitInfoMap.at(i_trk).at(irow);
				fillTupHit(i_trk, irow, hits, hitInfo);
  			}
  			catch(std::out_of_range&) {
				// ignore this exception
#ifdef KF_DEBUG
  	  			streamlog_out( DEBUG ) << "hitInfoMap: out of range: = " << i_trk 
					               << "   irow = " << irow 
						       << std::endl << std::flush;
#endif
  			}
		}
		
		// fill cluster data
  	  	for (int irow = 0; irow < NMAXROWS; irow++) {  // loop over rows
			try {
				auto &padInfoVec = padInfoMap.at(i_trk).at(irow);
				fillTupClr(i_trk, clrs, padInfoVec);
			} 
  			catch(std::out_of_range&) {
				// ignore this exception
  			}
    		} // end of row loop
  	} //end of track loop
}


/// Create KalHit from TrackerHit for KalTest.
EXTPCHit *RootFileProcessor3::createKalHit(const EVENT::TrackerHit *hitp, bool ismerged)
{
	auto &hit   = *hitp;
	auto  modID = getModuleID(hit) ; 
	auto  rowID = getRowID(hit, *_moduleVec[modID]);
	auto  hpos  = hit.getPosition();       // global frame
	TVector3 xv(hpos[0], hpos[1], hpos[2]); // 3d hit position vector

        auto superlayer   = (modID < 2 ? 0 : (modID < 5 ? 1 : 2 ));
        auto rowSerial    = superlayer * _moduleVec[modID]->getNRows() + rowID; 
        auto measlayerPtr = static_cast<EXTPCMeasLayer *>(_kaldetPtr->At(rowSerial));

	auto side = hpos[2] > 0 ? +1 : -1;
	auto mv   = measlayerPtr->XvToMv(xv, side);   // meas. vector (r*phi, z)
	auto d    = _maxDriftLength - side * mv(1, 0); // drift length
	double meas [2];
	double dmeas[2];
	meas [0]  = mv(0, 0);
	meas [1]  = mv(1, 0);
	dmeas[0]  = measlayerPtr->GetSigmaX(d) ;
	dmeas[1]  = measlayerPtr->GetSigmaZ(d) ;
#if 0

	auto qualityFlag = hit.getQuality();
	LCFlagImpl aflag(qualityFlag);

	// FIXME: //////////////////////////////////////////////////////////
	//    to be moved to more appropriate place
	int const SPLIT_HIT_BIT_INDEX = 20; // bit 20
	////////////////////////////////////////////////////////////////////
	if (hitflag::isMultipleHitCandidate(qualityFlag)) { // FIXME: very temporary
		if (!aflag.bitSet(SPLIT_HIT_BIT_INDEX)) dmeas[0] *= 1.e1;
		else				  	dmeas[0] *= 2.;
	}
#else
	if (ismerged) {
		auto covMat = hitp->getCovMatrix();
		auto drphi = sqrt(covMat[0] + covMat[2]);
		auto qualityFlag = hit.getQuality();
		LCFlagImpl aflag(qualityFlag);

		// FIXME: //////////////////////////////////////////////////////////
		//    to be moved to more appropriate place
		int const SPLIT_HIT_BIT_INDEX = 20; // bit 20
		////////////////////////////////////////////////////////////////////
		if (hitflag::isMultipleHitCandidate(qualityFlag)) { // FIXME: very temporary
			if (!aflag.bitSet(SPLIT_HIT_BIT_INDEX)) {
#if 0
		        	streamlog_out(DEBUG) << " createKalHit : multi hit: split hit? no : drphi = " << drphi << " dz = " << sqrt(covMat[5]) << endl;
#endif
			} else {
				dmeas[0]  = drphi;
#if 0
		        	streamlog_out(DEBUG) << " createKalHit : multi hit: split hit? yes : drphi = " << drphi << " dz = " << sqrt(covMat[5]) << endl;
#endif
			}
		}
	}
#endif
	// practically ignore hits on bad rows
        if (_badRowsOverride.end() != find(_badRowsOverride.begin(), _badRowsOverride.end(), rowID)) {
                dmeas[0] *= 100000.;
                dmeas[1] *= 100000.;
        }

        auto b = EXTPCKalDetector::GetBfield();
        auto v = EXTPCKalDetector::GetVdrift();
	auto kalhitp = new EXTPCHit(*measlayerPtr, meas, dmeas, side, v, hitp, b);

	return kalhitp;
}


/// Create KalTrack from TrackerHitVec by Kalman Filtering it.
EXHYBTrack *RootFileProcessor3::createKalTrack(TObjArray     &kalhits, // KalHit array
					       const bool     dir,     // filter direction
  	  				       map<int, int> &nhits)   // (row, #hits)
{		
	// Note:
	//        typedef std::vector<TrackerHit *> TrackerHitVec;
	//--
	// Choose 3 points to decide starting helix.
	//--
	int i1 , i2 , i3 ; // (i1,i2,i3) = (1st,mid,last) hit to filter
#if 1
	if (dir == kIterBackward) {
		i3 = 1 ; 
		i1 = kalhits.GetEntries() - 2 ;
		i2 = i1 /2;
	} else {
		i1 = 1 ;
		i3 = kalhits.GetEntries() - 2 ; 
		i2 = i3 / 2 ;
	}
#else
	if (dir == kIterBackward) {
		i3 = 0 ; 
		i1 = kalhits.GetEntries() - 1 ;
		i2 = i1 /2;
	} else {
		i1 = 0 ;
		i3 = kalhits.GetEntries() - 1 ; 
		i2 = i3 / 2 ;
	}
#endif

	//--
	// Create dummy site.
	//--
	auto  hit1p = static_cast<EXTPCHit *>(kalhits.At(i1));
	auto  hitdp = new EXTPCHit(*hit1p); // dummy hit
	auto &hitd  = *hitdp;
	      hitd(0,1) = 1.e6; // huge error on r*phi
	      hitd(1,1) = 1.e6; // huge error on z
	auto &sited = *new TKalTrackSite(hitd);
	sited.SetHitOwner(); // dummy site owns dummy hit
	sited.SetOwner();

	// decide starting helix using 3 points from kalhits
	auto &h1 = *static_cast<TVTrackHit *>(kalhits.At(i1));
	auto &h2 = *static_cast<TVTrackHit *>(kalhits.At(i2));
	auto &h3 = *static_cast<TVTrackHit *>(kalhits.At(i3));
	auto  x1 = h1.GetMeasLayer().HitToXv(h1);
	auto  x2 = h2.GetMeasLayer().HitToXv(h2);
	auto  x3 = h3.GetMeasLayer().HitToXv(h3);
	THelicalTrack helstart(x1, x2, x3, h1.GetBfield(), dir);

	// set dummy state vector
	static TKalMatrix svd(kSdim, 1);
	svd(0, 0) = 0.;
	svd(1, 0) = helstart.GetPhi0();
	svd(2, 0) = helstart.GetKappa();
	svd(3, 0) = 0.;
	svd(4, 0) = helstart.GetTanLambda();
	if (kSdim == 6) svd(5 , 0) = 0.; // t0
	static TKalMatrix C(kSdim, kSdim);
	for (int i = 0; i < kSdim; i++) C(i,i) = 1.e4 ; // dummy error matrix
	sited.Add(new TKalTrackState(svd, C, sited, TVKalSite::kPredicted));
	sited.Add(new TKalTrackState(svd, C, sited, TVKalSite::kFiltered));

	// create KalTrack
	auto  kaltrkp  = new EXHYBTrack; // a track is a kal system
	auto &kaltrack = *kaltrkp;
	kaltrack.SetOwner();  // track owns sites
	kaltrack.Add(&sited); // add dummy site

	auto nBadHits = 0;
	// Now start Kalman filter 
#ifndef KF_RETRY
	for (auto hitp : kalhits) { // loop over kalhits
		auto &kalhit = *static_cast<EXTPCHit *>(hitp);    // corresponding kalhit
#else
	TIter next(&kalhits, dir);
	EXTPCHit *hitp = nullptr;
	while ((hitp = static_cast<EXTPCHit *>(next()))) {
		auto &kalhit = *hitp;    // corresponding kalhit
#endif
#ifdef KF_EXP
		if (isMultiHit(kalhit) && !isSplitHit(kalhit)) {
			nBadHits++;
			continue;
		}
#endif
		auto &site   = *new TKalTrackSite(kalhit);

      		// FIXME //////////////////////////////////////////////////
      		// Temporary remove rows with >1 hits (no curling tracks)
      		int rowSerial = site.GetHit().GetMeasLayer().GetIndex();
      		if (nhits[rowSerial] > 1) {
      		  	delete &site;
      		  	continue;
      		}
      		///////////////////////////////////////////////////////////

		if (!kaltrack.AddAndFilter(site)) {
			streamlog_out(DEBUG2) << "Kalman Filter: site discarded!"
					      << " multi? " << (isMultiHit(kalhit) ? "yes" : "no " )
					      << " split? " << (isSplitHit(kalhit) ? "yes" : "no " )
					      << " rowSerial = " << rowSerial 
					      << "  mv(0) = " << kalhit.GetX(0)
					      << " dmv(0) = " << kalhit.GetDX(0)
					      << "  mv(1) = " << kalhit.GetX(1)
					      << " dmv(1) = " << kalhit.GetDX(1)
					      << endl << flush;
			kalhit.SetBit(kDiscarded); // mark this hit as discarded
			delete &site;
		}
	} // filterling ends here

	//--
	// Make sure that # discarded sites is small enough.
	//--
#ifndef KF_EXP
	auto frac = 1. - (kaltrack.GetEntries() - 1.)/kalhits.GetEntries();
#else
	auto frac = 1. - (kaltrack.GetEntries() - 1. + nBadHits)/kalhits.GetEntries();
#endif

	if (frac > 0.) {
		streamlog_out(DEBUG) << "createKalTrack: " << static_cast<void *>(kaltrkp) 
				     << " nBadHits = " << nBadHits
				     << " fraction discarded = " << frac << endl << flush;
	}

	// FIXME: //////////////////
#if 0
	auto kMaxBadFracCut = 0.2;
#else
	auto kMaxBadFracCut = 0.1;
#endif
	////////////////////////////
	if (frac > kMaxBadFracCut) {
		delete kaltrkp;
		kaltrkp = nullptr;
	}

	//--
	// That's it!
	//--
	return kaltrkp;
}

/// Calculate track's incident angle and expected hit position corresponding to the TrackerHit in question.
tuple<double, TVector3> RootFileProcessor3::getIncidentPhiAndPosition(EXHYBTrack      &kaltrack, 
		                                                const EXTPCHit        &kalhit, 
		                                                const gear::TPCModule &tpcmodule)
{
	// local utility lambda function
	auto toBaseRange = [](double phi)
	{
		while (phi <= -M_PI) {
			phi += 2. * M_PI ;
		}
		while (phi >   M_PI) {
			phi -= 2. * M_PI ;
		}
		return phi ;
	};

	// calculate radial vector(rv) at the hit point 
	
	auto &measlyr = kalhit.GetMeasLayer(); // TVMeasLayer
	auto pos   = measlyr.HitToXv(kalhit);  // in global frame (Cartesian)
	auto rphiv = tpcmodule.globalToLocal(pos.X(), pos.Y()); // in module-local frame (polar)
	auto r     = rphiv[0];
	auto phi   = rphiv[1];
	TVector2 rv2(r*cos(phi), r*sin(phi));

	// then calculate tangential vector(tv) at the hit point 
	auto &cursite  = kaltrack.GetCurSite();
	auto &curstate = static_cast<TKalTrackState &>(cursite.GetCurState());
        auto  heltrack = curstate.GetHelix();

	TVector3 xvexpect(pos.X(), pos.Y(), pos.Z());
	auto &surface  = dynamic_cast<const TCylinder &>(measlyr);		        // TCylinder
	auto  dphi     = 0.;							        // deflection angle from pivot
	auto  mode     = 0;
	heltrack.MoveTo(xvexpect, dphi);					        // move pivot to the hit point
	      dphi     = 0.;
	auto  nxp      = surface.CalcXingPointWith(heltrack, xvexpect, dphi, mode);     // calculate expected hit point
        if (nxp != 0) {
		auto  globaltv = heltrack.CalcDxDphi(dphi);                             // tangential vector in global frame (Cartesian)
		auto  rphitv   = tpcmodule.globalToLocal(globaltv(0,0), globaltv(1,0)); // tangential vector in module-local frame (polar)
        	auto  rtv      = rphitv[0];
        	auto  phitv    = rphitv[1];
		TVector2 tv2(rtv*cos(phitv), rtv*sin(phitv));                           // tangential vector in module-local frame (Cartesian)
		tv2 = tv2.Unit();

		if (heltrack.GetKappa() > 0) tv2 *= -1. ;

		// calculate incident angle(phi) as angle between tv2 and rv2
		dphi = rv2.DeltaPhi(tv2);
		dphi = toBaseRange(dphi);
		if (abs(dphi) > 0.5*M_PI) dphi = dphi > 0. ? (dphi - M_PI) : (dphi + M_PI);
	} else { // could not find any crossing point. 
		dphi = 0.;
		xvexpect.SetXYZ(pos.X(), pos.Y(), pos.Z()); //FIXME: do something more sensible here.

		streamlog_out(ERROR) << "found no crossing point" << endl;
	}

	return {dphi, xvexpect}; // (phi_inc, xv_expected)
}


/// Calculate track's expected hit position at a given row
TVector3 RootFileProcessor3::getExpectedHitPosition(const EVENT::Track &trk, int serialrow)
{
	// local utility lambda function
	auto toBaseRange = [](double phi)
	{
		while (phi <= -M_PI) {
			phi += 2. * M_PI ;
		}
		while (phi >   M_PI) {
			phi -= 2. * M_PI ;
		}
		return phi ;
	};

	//--
	// Prepare a helix for input track.
	//--
	// FIXME: this should be somewhere more appropriate ////////////////////////////
	static const Double_t kLightVelocity = 2.99792458e8; //! light velocity [m/sec]
	static const Double_t kGiga          = 1.0e9;        //! Giga = 10^{9}
	////////////////////////////////////////////////////////////////////////////////

	auto bz      =  EXTPCKalDetector::GetBfield();
	auto phi0    =  toBaseRange(trk.getPhi() - M_PI / 2.);
	auto ptor    =  kGiga/kLightVelocity*1000./bz; // units: mm, sec, Tesla
	auto kappa   =  ptor * trk.getOmega();
	auto dr      = -trk.getD0();
	auto dz      =  trk.getZ0();
	auto tanl    =  trk.getTanLambda();
	auto refpos  =  trk.getReferencePoint();
	auto x0      =  double(refpos[0]);
	auto y0      =  double(refpos[1]);
	auto z0      =  double(refpos[2]);

	THelicalTrack heltrack(dr, phi0, kappa, dz, tanl, x0, y0, z0, bz);

	//--
	// Get measurement layer for input row.
	//--
        auto &measlyr  = *static_cast<EXTPCMeasLayer *>(_kaldetPtr->At(serialrow));

	//--
	// Calculate the expected hit point for this row.
	//--
	TVector3 xvexpect;
	auto &surface  = dynamic_cast<const TCylinder &>(measlyr);		    // TCylinder
	auto  dphi     = 0.;							    // deflection angle from pivot
	auto  mode     = 0;
	auto  nxp      = surface.CalcXingPointWith(heltrack, xvexpect, dphi, mode); // calculate expected hit point
        if (nxp == 0) {
		xvexpect.SetXYZ(-1.e6, -1.e6, -1.e6); //FIXME: do something more sensible here.

		streamlog_out(ERROR) << "found no crossing point" << endl;
	}

	return xvexpect;
}
