#include "EventMergerProcessor.h"

#include <iostream>

#include "IO/LCReader.h"
#include "IO/LCWriter.h"

#include "EVENT/LCEvent.h"
#include "EVENT/LCCollection.h"
#include "EVENT/Track.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackerRawDataImpl.h"
#include "IMPL/LCFlagImpl.h"
#include "UTIL/LCTOOLS.h"

#include "marlin/Exceptions.h"
// ----- include for verbosity dependend logging ---------
#include "marlin/VerbosityLevels.h"

using namespace lcio ;
using namespace marlin ;

namespace eventmerger {
  
    /// an instance of this processor created upon shared lib loading
    EventMergerProcessor aEventMergerProcessor;
    
  
    /// C-tor
    EventMergerProcessor::EventMergerProcessor() : Processor("EventMergerProcessor"), 
 	       _lcReader(0),
  	       _nRun(0),
  	       _nEvt(0),
  	       _nMergedEvt(0)
    {
        // modify processor description
        _description = "Merge every consecutive two events..." ;
    
        // register steering parameters: name, description, class-variable, default value
        registerInputCollection(LCIO::TRACK,
                                "InputTrackCollectionName" ,
                                "Name of the input track collection"  ,
                                _inputTrackColName ,
                                std::string("TPCTracks"));
  
        registerInputCollection(LCIO::TRACKERRAWDATA,
                                "InputCollectionName" ,
                                "Name of the input collection"  ,
                                _inputColName ,
                                std::string("TPCRawData"));

        registerOutputCollection(LCIO::TRACKERRAWDATA,
                                 "OutputCollectionName" ,
                                 "Name of the output collection"  ,
                                 _outputColName ,
                                 std::string("MergedTPCRawData")
                                );

        registerOptionalParameter("MinimumNumberOfTracks", "minimum number of tracks", _nTracksMin, int(1));

        registerOptionalParameter("MaximumNumberOfTracks", "maximum number of tracks", _nTracksMax, int(1));

        registerOptionalParameter("MinimumNumberOfDegreesOfFreedom", "minimum number of degrees of freedom", _nDfMin, int(40));
    }
  
  
    /// job initialization
    void EventMergerProcessor::init()
    { 
        streamlog_out( DEBUG4 ) << "***** EventMergerProcessor::init() *****" << std::endl << std::flush;
  
        // usually a good idea to
        printParameters() ;
    
        // opening background input
        _lcReader = LCFactory::getInstance()->createLCReader(LCReader::directAccess) ;
    
        // initialize run No. and event No.
        _nRun = 0 ;
        _nEvt = 0 ;
        _nMergedEvt = 0 ;
    }
  
  
    /// run initialization
    void EventMergerProcessor::processRunHeader([[maybe_unused]] LCRunHeader* run)
    { 
        streamlog_out( DEBUG4 ) << "***** EventMergerProcessor::processRunHeader(...) *****" << std::endl << std::flush;
        _nRun++ ;
    } 
  
  
    /// process event
    ///   Note:
    ///      after this processor, 
    ///         odd  output events contain only unmerged events, while
    ///         even output events contain both merged events as well as unmerged ones.
    ///
    void EventMergerProcessor::processEvent(LCEvent * evtp)
    {
        streamlog_out( DEBUG4 ) << std::endl
                                << "***** EventMergerProcessor::processEvent(...) *****" << std::endl
                                << "   Event No: " << _nEvt                              << std::endl
                                << "***************************************************" << std::endl
                                << std::flush;
  
        // increment event counter
        _nEvt++;

        // get track collection
        EVENT::LCCollection* curtrkcolp = nullptr;
        try {
            curtrkcolp = evtp->getCollection(_inputTrackColName);
            // make sure the number of good tracks is within the allowed range
	    auto nTracks = curtrkcolp->getNumberOfElements();
	    // count # good tracks
	    auto ntrk    = 0;
	    for (int itrk = 0; itrk < nTracks; itrk++) {
                auto trkp = static_cast<EVENT::Track *>(curtrkcolp->getElementAt(itrk));
                auto ndf  = trkp->getNdf();
                if (ndf >= _nDfMin) ntrk++;
	    }
            // require # good tracks is within the allowed range
            if (ntrk < _nTracksMin || ntrk > _nTracksMax || nTracks > _nTracksMax) {
                StopProcessingException skip(this);
		if (ntrk < _nTracksMin || ntrk > _nTracksMax) {
                    streamlog_out( WARNING ) << "The number of good TPC tracks, " << ntrk << ", is out of range: [" 
                                             << _nTracksMin << ", " << _nTracksMax << "]!  Skipped."
                                             << std::endl << std::flush;
		} else {
                    streamlog_out( WARNING ) << "The number of TPC tracks, " << nTracks << ", is out of range: [" 
                                             << _nTracksMin << ", " << _nTracksMax << "]!  Skipped."
                                             << std::endl << std::flush;
		}
                return;
            }

            streamlog_out( DEBUG4 ) << "Event accepted."
                                    << std::endl << std::flush;
        } catch (std::exception&) {
            // no track collection
            StopProcessingException skip(this);
            streamlog_out( WARNING ) << "No track collection! Skipped."
                                     << std::endl << std::flush;
            return;
        }
  
        // get target collection to merge
        EVENT::LCCollection* curcolp = nullptr;
        try {
            curcolp = evtp->getCollection(_inputColName);
            // make sure it's not empty
            if (!curcolp->getNumberOfElements()) {
                StopProcessingException skip(this);
                streamlog_out( WARNING ) << "Empty TPC raw data collection! Skipped."
                                         << std::endl << std::flush;
                return;
            }
        } catch (std::exception&) {
            //std::cerr << "This event has no TPC raw data collection ! Skipped." << std::std::endl;
            StopProcessingException skip(this);
            streamlog_out( WARNING ) << "No TPC raw data collection! Skipped."
                                     << std::endl << std::flush;
            return;
        }
  
        streamlog_out( DEBUG4 ) << "   started processing this event" << std::endl << std::flush;
  
        // set the flags that cellID1 should be stored
        lcio::LCFlagImpl rdatFlag(0) ;
        rdatFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
  
        // create an output collection
        LCCollection* outputCol = new LCCollectionVec(LCIO::TRACKERRAWDATA);
        outputCol->setFlag(rdatFlag.getFlag());
  
        static LCCollectionVec *prevcolp = nullptr; // buffer to keep previous event
  
        if (!prevcolp) { // event buffer empty
            // create a buffer
            prevcolp = new LCCollectionVec(LCIO::TRACKERRAWDATA);
            prevcolp->setFlag(rdatFlag.getFlag());
  
            // copy *curcolp to *prevcolp
            auto &objvec = *static_cast<LCCollectionVec *>(curcolp);
            for (auto objp : objvec) {
                auto srcp = static_cast<TrackerRawDataImpl *>(objp);
                auto cpp  = new TrackerRawDataImpl(*srcp); //FIXME: check if default C-tor works.
                prevcolp->addElement(cpp);
            }
  
            streamlog_out( DEBUG4 ) << "   copied this event to a buffer" << std::endl << std::flush;
  
            // add empty output collection
            evtp->addCollection(outputCol, _outputColName);
  
        } else { // found data to merge in the buffer
            // merge previous event to the current one
            merge(prevcolp, curcolp, outputCol);
  
            streamlog_out( DEBUG4 ) << "   merged this event with the one in the buffer" << std::endl
                                    << std::flush;
  
            // clean up the buffer
            delete prevcolp;
            prevcolp = nullptr;
  
            // add the result to output collection
            evtp->addCollection(outputCol, _outputColName);
  
            streamlog_out( DEBUG4 ) << "   added the merged event to output collection" << std::endl
                                    << "     --------------------------------"          << std::endl
                                    << "     merged Event No: " << _nMergedEvt          << std::endl
                                    << "     --------------------------------"          << std::endl
                                    << std::flush;
  
            _nMergedEvt++; 
        }
    }
  
  
    /// event check
    void EventMergerProcessor::check([[maybe_unused]] LCEvent * evtp)
    { 
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }
  
  
    /// job end
    void EventMergerProcessor::end()
    { 
        streamlog_out( DEBUG4 )  << std::endl  
                                 << "***** EventMergerProcessor::end() *****" << std::endl << std::flush;
    
        streamlog_out( MESSAGE ) << " ---------------------------------------------------- " << std::endl
    			         << "   EventMergerProcessor processor " << _nMergedEvt 
    			         << " events created with " << _nEvt << " input events."     << std::endl
    			         << std::flush;
    }

    using namespace std;

    /// fill ADC data in a channel-to-ADC-data map
    void EventMergerProcessor::fillADC (LCCollection                        *colp,    // TrackerRawData collection
                                        map<long long, vector<short int> *> &ADCMap)  // channel to adc data map
    {   
        ///// FIXME: should be somewhere else ////////////////////////
        static const int kMaxTimeBin = 1024; // maximum time bin value
        //////////////////////////////////////////////////////////////
    
        //--
        // Loop over channels and fill a channel-to-ADC-data map
        //--
        int nElements = colp->getNumberOfElements();
        for (int i=0; i<nElements; i++) {
            auto rawdatap = dynamic_cast<TrackerRawDataImpl *>(colp->getElementAt(i));
    	    if (!rawdatap) continue;
    
    	    auto longID = cellID2long(rawdatap->getCellID0(), rawdatap->getCellID1()); // (long long) channel ID
    	    vector<short int> * adcvecp = ADCMap[longID];
    	    if (!adcvecp) { // create a new adc buffer for this channnel if not already there
    	        adcvecp = new vector<short int>(kMaxTimeBin, 0.);
    	        ADCMap[longID] = adcvecp;
    	    }
    	    auto &adcvec = *adcvecp;
    	    
    	    const EVENT::ShortVec & adcv = rawdatap->getADCValues(); 	
    	    auto  nadc  = adcv.size();
    	    auto  tbin0 = rawdatap->getTime();
    
  	    for (unsigned int iadc = 0; iadc < nadc; iadc++) {
  	        auto itime = tbin0 + iadc;
  	        if (itime < 0 || itime >= kMaxTimeBin ) { // check if this is in allowed time range
#if 0
                    streamlog_out( WARNING ) << "Merger::fillADC: out of range error: itime: " << itime 
              		                     << endl 
              			             << "   ignore this time slot! !"
              			             << endl << flush;
#endif
                } else { 
                    adcvec[itime] = adcvec[itime] + adcv[iadc]; // add this adc value to this time bin
                }
            }
        }
    }


    /// merge two collections
    void EventMergerProcessor::merge(LCCollection * src1p, // 1st source collection
    		                 LCCollection * src2p, // 2nd source collection
    		                 LCCollection * outp)  // output merged collection
    {
        ///// FIXME: should be somewhere else /////////////
        static const short int kThreshold = 1; 
        static const short int kPedADC    = 100; 
        static const short int kMaxADC    = 1024 - kPedADC; 
        ///////////////////////////////////////////////////
        
        //--
        // Check if collections have the same type
        //--
        auto src2Type = src2p->getTypeName();
        if (src2Type != src1p->getTypeName()) {
            streamlog_out( WARNING ) << "merge not possible, collections of different type" << endl << flush;
            return;
        }
        streamlog_out( DEBUG4 ) << "merging collection of type: " << src2Type << " --- " << endl << flush;
        
        //--
        // Merge TPC raw data
        //--
        if (src2Type == LCIO::TRACKERRAWDATA) {
            auto nElementsSrc1 = src1p->getNumberOfElements();
            auto nElementsSrc2 = src2p->getNumberOfElements();
            map<long long, vector<short int> *> ADCMap;
    
            streamlog_out( DEBUG4 ) << "merging ... src1 with " << nElementsSrc1 << " elements"
    		                    << " with src2 with " << nElementsSrc2 << " elements" << endl << flush;
          
            fillADC(src1p, ADCMap);
            fillADC(src2p, ADCMap);
    
            auto nChannels = 0;
            for (auto [longID, adcvecp] : ADCMap) { // loop over channels
                auto id0 = cellID0fromlongID(longID);
                auto id1 = cellID1fromlongID(longID);
                nChannels++;
#if 0
                streamlog_out ( DEBUG5 ) << "-------------------------------------------- "<< endl
                                         << "nChannels: " << nChannels                     << endl
                                         << "longID "     << (void *) longID               << endl
                                         << "id0: "       << id0 << " id1: " << id1        << endl
                                         << "adcvecp: "   << (void *) adcvecp              << endl
                                         << "-------------------------------------------- "<< endl
                                         << flush;
#endif
      	        auto &adcvec  = *adcvecp;  // vector<short int> &
                EVENT::ShortVec adcvalues; // output adc buffer	
  
  	        int  nadc      = adcvec.size();
  	        int  itimebin0 = -9999;
   	        bool isfound   = false;
  
  	        for (int itime = 0; itime < nadc; itime++) { // loop over time bins
  	            short int adc = adcvec[itime];
  	  
  	            if (!isfound && adc < kThreshold) {         // below threshold, yet no pulse
  	                continue;
  	            } else if (!isfound && adc >= kThreshold) { // a pulse found
  	                isfound   = true;
                        itimebin0 = itime;
                        adcvalues.clear();                      // clear output adc buffer
                        adc = min(adc, kMaxADC);                // adc saturation
                        adcvalues.push_back(adc);               // store this adc value to the buffer
  	            } else if (isfound && adc >= kThreshold) {  // pulse continues
                        adc = min(adc, kMaxADC);
                        adcvalues.push_back(adc);
                    } else if (isfound && adc < kThreshold) {   // end of this pulse
                        isfound = false;
    		    // store this pulse in the output collection
                        TrackerRawDataImpl * rawdatap = new TrackerRawDataImpl;
                        rawdatap->setADCValues(adcvalues);
                        rawdatap->setTime(itimebin0);
                        rawdatap->setCellID0(id0);
                        rawdatap->setCellID1(id1);
                        outp->addElement(rawdatap);
                    }
                }
                delete adcvecp;
                ADCMap[longID] = 0;
            } 
    
            streamlog_out ( DEBUG5 ) << "-------------------------------------------- "<< endl
                                     << "Merged  " << nChannels << " channels "        << endl
                                     << "-------------------------------------------- "<< endl
                                     << flush;
        }
    }
  
} // end namespace 
