#define FLAT_DIST
#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCRunHeaderImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/MCParticleImpl.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>
#include <typeinfo>
#include <cmath>
#include <cstdlib>

#include <CLHEP/Random/RandGauss.h>
#include <CLHEP/Random/RandFlat.h>

using namespace std ;
using namespace lcio ;

int main(int argc, char **argv)
{
  //--
  // Parse input arguments.
  //--
  unsigned int nEvents     = 100;
  [[maybi_unused]] 
  double       yoffset_in  = 4.;    // [mm]
  double       zpos        = 270.;  // [mm]
  double       energy_in   = 5.;    // [GeV]
  double       phi_in      = 0.;    // [deg]
  double       lambda_in   = 0.;    // [deg]
  string       outFileName = "electron5GeV.twotrack.slcio";

#if 0
  double       sigphi      = 0.1;   // [rad]
  double       siglamb     = 0.002; // [rad]
  double       xpos        = 1439.; // [mm]
#else
  double       sigphi      = 0.01;  // [rad]
  double       siglamb     = 0.002; // [rad]
#if 0
  double       xpos        = 1200.; // [mm]
#else
  double       xpos        = 1439.; // [mm]
#endif
#endif
  double       ypos        = 50.;   // [mm]
  double       sigx        =  0.;   // [mm]
#ifndef FLAT_DIST
  double       sigy        =  5.;   // [mm]
  double       sigz        =  1.;   // [mm]
#else
  double       sigy        =  5.;   // [mm]
  double       sigz        =  5.;   // [mm]
#endif
  double       sigp        =  1.;   // [GeV]


  switch (argc) {
    case (8) :
	    outFileName = argv[7];
	    [[fallthrough]];
    case (7) :
	    lambda_in   = atof(argv[6]);
	    [[fallthrough]];
    case (6) :
	    phi_in      = atof(argv[5]);
	    [[fallthrough]];
    case (5) :
	    energy_in   = atof(argv[4]);
	    [[fallthrough]];
    case (4) :
	    zpos        = atof(argv[3]);
	    [[fallthrough]];
    case (3) :
	    yoffset_in  = atof(argv[2]);
	    [[fallthrough]];
    case (2) :
	    nEvents     = atof(argv[1]);
	    [[fallthrough]];
    case (1) :
	    break;
    default:
	    cout << "Usage: "
		 << "     ./generateElectronPairsForLP1 [nev [ee [z [phi [lamb [fname]]]]]] "
		 << "   where "
		 << "      nev   : # events to generate     "
		 << "      ee    : electron energy     [GeV]"
		 << "      z     : starting z position [mm] "
		 << "      phi   : starting phi angle  [deg]"
		 << "      lamb  : lambda              [deg]"
		 << "      fname : output file name         "
		 << endl;
	    return 1;
  }
  
  static const double m = 0.000511; // the electron mass in GeV

  if (energy_in < m) {
    cout << "Error, energy cannot be smaller than mass of the electron ( "
	 << m << " GeV)!" << endl;
    return -1;
  }

  //--
  // Open outfile			     
  //--
  // create sio writer
  LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
  
  try{  lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
  }
  catch( IOException& e ){    
    cout << e.what() << endl ;
    return 0 ;
  }
      
  //--
  // Create and write runheader
  //--
  LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
  int runNr = 0;
  runHdr->setRunNumber( runNr) ;
  runHdr->setDetectorName ("LPTPC");      
  lcWrt->writeRunHeader( runHdr ) ;
  
  //--
  // Loop over events
  //--
  for (unsigned int eventNr = 0; eventNr < nEvents; eventNr++) {
    //--
    // Create new lcio event
    //--
    LCEventImpl*  evt = new LCEventImpl() ;
    
    evt->setRunNumber(  runNr   ) ;
    evt->setEventNumber( eventNr ) ;
  
    // we want to store a collection of trackerRawData
    LCCollectionVec* mcElectronsVec = new LCCollectionVec( LCIO::MCPARTICLE )  ;

    //--
    // Loop over 2 electrons
    //--
    double phi0 = phi_in   *M_PI/180 + CLHEP::RandGauss::shoot(0., sigphi );
    for (int ie = 0; ie < 2; ie++) {
      // preparation: calculate the momentum
      double lambda = lambda_in*M_PI/180 + CLHEP::RandGauss::shoot(0., siglamb);
      double p_in   = sqrt((energy_in - m)*(energy_in + m));
      double ap     = p_in + CLHEP::RandGauss::shoot(0., sigp);

      double phi = phi0*(1-2*ie);
      double p[3] = { ap * cos(phi) *cos(lambda),
                      ap * sin(phi) *cos(lambda),
                      ap * sin(lambda) };

      //beam spread of 1mm in all directions except z to avoid binning effects in simulation
      double randx = CLHEP::RandGauss::shoot( 0., sigx );
#ifndef FLAT_DIST
      double randy = CLHEP::RandGauss::shoot( 0., sigy );
      double randz = CLHEP::RandGauss::shoot( 0., sigz );

      double vtx[3] = {xpos+randx, ypos+yoffset_in*(phi > 0 ? -1. : +1.)+randy, zpos+randz};
#else
      double randy = CLHEP::RandFlat::shoot( -sigy, sigy );
      double randz = CLHEP::RandFlat::shoot( -sigz, sigz );

      double vtx[3] = {xpos+randx, ypos+randy, zpos+randz};
#endif

      int    pid    = 11*(1-2*ie);
      double charge = -(1.-2.*ie);

      // create MC particle
      MCParticleImpl *electron =  new MCParticleImpl;
      electron->setPDG(pid);           // it's an electron
      electron->setCharge(charge);     // electrons have chage -1
      electron->setMass( m );          // the electron mass
      electron->setGeneratorStatus(1); // a stable particle in the detector
      electron->setMomentum( p );      // the the corresponding momentum 
      electron->setVertex(vtx);

      // add this electron to collection
      mcElectronsVec->addElement( electron );
    }

    // add the rawData collection to the event
    evt->addCollection( mcElectronsVec , "MCElectrons" ) ;

    // write the event to the file
    lcWrt->writeEvent( evt ) ;
    
    // dump the event to the screen 
    LCTOOLS::dumpEvent( evt ) ;

    // ------------ IMPORTANT ------------- !
    // we created the event so we need to delete it ...
    delete evt ;
    // -------------------------------------
  } // for eventNr
  
  lcWrt->close() ;
     
  return 0;
}
